"""
Exceptions
^^^^^^^^^^

Exceptions used by the quality module.
"""

################################
# Exceptions
################################


class NotInitialized(Exception):
    """Initialize to discover all available expectation
    implementations. It has not happened.
    """

    pass


class InvalidExpectation(Exception):
    """
    Invalid specification of expectations
    """

    pass


class NoExpectations(Exception):
    """
    Empty list of expectations to evaluate
    """

    pass


class UnsupportedExpectation(Exception):
    """
    Expectation name or content not implemented/supported.
    """

    pass


class IncorrectImplementationExpectation(Exception):
    """
    Expectation implementation class didnt return right result
    """

    pass


class InvalidConfigurationExpectation(Exception):
    """
    Expectation implementation class didnt return right result
    """

    pass



import os
import sys
import json
import copy
import re
import traceback
import logging
import markdown
from datetime import timezone
import pytz
import humanize
from ..lib.exceptions import *
from .node import *

logger = logging.getLogger("app")

__all__ = ["SearchSkin", "GenericSkin"]


class SearchSkin(Skin):
    """

    This provides a search interface to the dataframes.

    .. deprecated::
           Use enrichsdk.render functions from the app

    """

    def __init__(self, *args, **kwargs):
        super(SearchSkin, self).__init__(*args, **kwargs)

        self.framename = None
        """
        Dataframe to service the search interface 
        """

        self.autocomplete_column = None
        """
        Column to feed the autocomplete function of the search interface
        """

        self.dashboard_rules = []
        """
        Specification of the layout of the search result. 
        """
        self.run = None
        self.detail = None
        self.tags.extend(["skin", "search"])

    def load(self):

        # If not loaded or loaded run not the last one:
        lastrun = self.config.runmanager.get_last_run()
        if (
            (self.run is not None)
            and (lastrun is not None)
            and (lastrun.runid == self.run.runid)
        ):
            return

        # Note it
        self.run = lastrun

        # If still not loaded, give up
        if self.run is None:
            return

        # Load the details...
        self.run.load_files(include_output_frame=True)
        self.detail = self.run.find_file_detail(
            framename=self.framename, action="output"
        )

        if self.detail is None:
            raise Exception("Could not find detail")

    def autocomplete(self, term):

        if self.run is None:
            try:
                self.load()
            except:
                pass

        if self.run is None or self.detail is None:
            return []

        df = self.detail["df"]
        names = df[self.autocomplete_column].values
        names = [n for n in names if term.lower() in n.lower()]

        data = [{"label": n, "key": n} for n in names][:10]

        return data

    def render_helper(self, rules, row):
        data = []
        html = ""
        for r in rules:
            enable = r.get("enable", True)
            if not enable:
                continue
            c = r["context"](row)
            rendered = self.template_render(r["template"], c)
            html += "\n" + rendered

        data.append({"target": "result", "rendered": html})

        return data

    def render(self, name):

        if self.run is None:
            try:
                self.load()
            except:
                pass

        if self.run is None:
            return {}

        if name is None or len(name) == 0:
            return {}

        if self.detail is None:
            # Try loading
            try:
                self.load()
            except:
                pass

        if self.detail is None:
            return {}

        df = self.detail["df"]
        row = df[df[self.autocomplete_column] == name]

        if len(row) > 0:
            row = row.iloc[0].to_dict()
            data = self.render_helper(self.dashboard_rules, row)
        else:
            data = []

        return data


class GenericSkin(Skin):
    """
    This skin allows rendering of arbitrary content.

    It takes all available dataframes in the last available run as
    input. The render function can display content from one or more
    dataframes.
    """

    def __init__(self, *args, **kwargs):
        super(GenericSkin, self).__init__(*args, **kwargs)
        self.run = None
        self.tags.extend(["skin", "generic"])

    def load(self):
        """
        Load the last run of the pipeline that this skin is
        included in.

        """
        # If not loaded, load now
        lastrun = self.config.runmanager.get_last_run()
        if (
            (self.run is not None)
            and (lastrun is not None)
            and (lastrun.runid == self.run.runid)
        ):
            return

        # Note it
        self.run = lastrun

        # If still not loaded, give up
        if self.run is None:
            logger.error("Could not find run in GenericSkin")
            return

        # Load the details...
        self.run.load_files(include_output_frame=True)

    def discover_params_helper(self, data):
        """
        This is overridden by the renderer
        """
        raise Exception("To be overridden by the renderer")

    def discover_params(self):
        """
        Parameters that can be passed

          {
             name: [val1, val2, ..]
          }

        """

        if self.run is None:
            try:
                self.load()
            except:
                traceback.print_exc()
                pass

        if self.run is None:
            raise Exception("Run not available")

        # Collect all the frames...
        data = self.run.get_file_hash()

        return self.discover_params_helper(data)

    def render_helper(self, rules, params):
        """
        This function renders the last run's data according
        to rules.

        Args:
            rules (list): A list of widget configurations

        Returns:

            Rendered html

        """
        # Collect all the frames...
        data = self.run.get_file_hash()

        response = []
        html = ""
        for r in rules:
            enable = r.get("enable", True)
            meta = r.get("meta", False)
            if not enable:
                continue

            t = r["template"]
            if not meta:
                c = r["context"](data, params)
                rendered = self.template_render(t, c)
                html += "\n" + rendered
            else:

                # Render the subset of rules...
                subresponse = self.render_helper(r["rules"], params)
                subresponsehtml = "\n".join([s["rendered"] for s in subresponse])

                # Now incorporate it...
                c = r["context"](data, params, subresponsehtml)
                rendered = self.template_render(t, c)
                html += "\n" + rendered

        response.append({"target": "result", "rendered": html})

        return response

    def render(self, params):
        """
        Generic render function that first loads the run,
        all the files that are part of the run, and call
        the render helper.
        """
        try:
            self.load()
        except:
            traceback.print_exc()
            pass

        if self.run is None:
            return {}

        data = self.render_helper(self.dashboard_rules, params)

        return data

    def get_metadata_context(self, data, params):
        """
        Include standard metadata
        """

        pairs = [
            {"name": "RunID", "value": self.run.runid},
            {
                "name": "Run Date",
                "value": humanize.naturaltime(self.run.start_time.replace(tzinfo=None)),
            },
        ]

        return {"title": "Run Metadata", "width": 12, "pairs": pairs}



import os
import copy
import json
import logging
from datetime import datetime
from dateutil.relativedelta import relativedelta
from django.urls import reverse
import sqlite3

from enrichsdk.lib.customer import find_usecase

from .models import ComplexSearchRecord

logger = logging.getLogger("app")

dbpath = os.path.expandvars("path-to-db.sqlite")


def get_complex_task_url(spec, request):
    return reverse("APPNAME:complex_request")


def get_filters():
    """
    All the Filters that must could be applied
    """
    if not os.path.exists(dbpath):
        return []

    # First get all the rules...
    con = sqlite3.connect(dbpath)
    cur = con.cursor()

    # => Get all the rules...
    table = "some_table"
    rules = list(cur.execute(f"SELECT distinct rule FROM {table} order by rule"))
    rules = [r[0] for r in rules]
    rule_conditions = []
    for index, r in enumerate(rules):
        rule_conditions.append({"value": r, "condition": f"(rule = '{r}')"})

    result = [{"name": "Matching Rule", "conditions": rule_conditions}]

    con.close()

    return result


def get_queries():

    base = """\
SELECT *
FROM {}
WHERE (%(conditions)s)
ORDER BY total_txn_value desc
limit {}"""

    # Note: We are implementing uniqueness of the kycid using the groupby
    return [
        {
            "name": "All Available",
            "sql": base.format("tablename", 50000),
            "clean": lambda df: df.groupby("primary_key_id").first().reset_index(),
        },
        {
            "name": "Top 10",
            "sql": base.format("tablename", 100),
            "clean": lambda df: df.groupby("primary_key_id").first().reset_index()[:10],
        },
        {
            "name": "Top 50",
            "sql": base.format("tablename", 500),
            "clean": lambda df: df.groupby("primary_key_id").first().reset_index()[:50],
        },
    ]


search_spec = {
    "name": "Profiles",
    "description": "Profiles for Multiple Usecases",
    "usecase": find_usecase(__file__),
    "links": [{"name": "APPNAME", "url": "APPNAME:index"}],
    "models": {"search_record": ComplexSearchRecord},
    "personas": [
        {
            "type": "sqlite",
            "dbpath": dbpath,
            "name": "Profiles",
            "description": "Profiles match one of many rules. Note that there could be multiple entries per",
            "tables": [
                {
                    "name": "Sample",
                    "table": "tablename",
                    "action": "sample",
                    "download": True,
                },
                {
                    "name": "Queries",
                    "table": "tablename",
                    "action": "query",
                    "download": True,
                    "queries": get_queries(),
                    "filters": get_filters(),
                    "workflow_columns": ["rule", "columns_to_post_to_workflow"],
                },
                {
                    "name": "Search",
                    "table": "tablename",
                    "action": "search",
                    "search_columns": ["columns_to_search"],
                    "workflow_columns": ["rule", "columns_to_post_to_workflow"],
                },
            ],
            "workflows": [
                {"name": "ComplexTask", "url": get_complex_task_url},
            ],
        },
    ],
}


def get_spec():
    spec = copy.copy(search_spec)
    return spec



import os
from prompt_toolkit.validation import Validator, ValidationError

#################################################
# Validators
#################################################
class NameValidator(Validator):
    def validate(self, document):
        if len(document.text) == 0:
            raise ValidationError(
                message="Input should not be empty", cursor_position=len(document.text)
            )  # Move cursor to end of input.


class RealtimeBatchValidator(Validator):
    def validate(self, document):
        if len(document.text) == 0:
            raise ValidationError(
                message="Computation should not be empty",
                cursor_position=len(document.text),
            )
        if document.text not in ["batch", "realtime"]:
            raise ValidationError(
                message="Valid values are 'batch' and 'realtime'",
                cursor_position=len(document.text),
            )


class EmailValidator(Validator):
    def validate(self, document):
        if len(document.text) == 0:
            raise ValidationError(
                message="Email should not be empty", cursor_position=len(document.text)
            )
        if "@" not in document.text:
            raise ValidationError(
                message="Invalid email address", cursor_position=len(document.text)
            )


class OrgDescValidator(Validator):
    def validate(self, document):
        if len(document.text) == 0:
            raise ValidationError(
                message="Customer organization description should not be empty",
                cursor_position=len(document.text),
            )  # Move cursor to end of input.


class GenericDescValidator(Validator):
    def validate(self, document):
        if len(document.text) == 0:
            raise ValidationError(
                message="Customer organization description should not be empty",
                cursor_position=len(document.text),
            )  # Move cursor to end of input.


class LogoValidator(Validator):
    def validate(self, document):
        url = document.text
        if len(url) == 0:
            raise ValidationError(
                message="Customer logo URL should not be empty",
                cursor_position=len(document.text),
            )  # Move cursor to end of input.
        if not url.startswith("http"):
            raise ValidationError(
                message="Customer logo URL should be a valid URL (http://...",
                cursor_position=len(document.text),
            )  # Move cursor to end of input.
        if not (
            url.lower().endswith("png")
            or url.lower().endswith("gif")
            or url.lower().endswith("jpg")
            or url.lower().endswith("ico")
        ):
            raise ValidationError(
                message="Customer logo URL should be a valid URL (http://....png",
                cursor_position=len(document.text),
            )  # Move cursor to end of input.


class URLValidator(Validator):
    def validate(self, document):
        url = document.text
        if len(url) == 0:
            raise ValidationError(
                message="URL should not be empty", cursor_position=len(document.text)
            )  # Move cursor to end of input.
        if not url.startswith("http"):
            raise ValidationError(
                message="URL should be a valid URL (http://...",
                cursor_position=len(document.text),
            )  # Move cursor to end of input.


class ValidDir(Validator):
    def validate(self, document):
        if (
            (len(document.text) == 0)
            or (not os.path.exists(document.text))
            or (not os.path.isdir(document.text))
        ):
            raise ValidationError(
                message="Directory specified doesnt exist or is invalid",
                cursor_position=len(document.text),
            )


class ValidFile(Validator):
    def validate(self, document):
        if (
            (len(document.text) == 0)
            or (not os.path.exists(document.text))
            or (not os.path.isfile(document.text))
        ):
            raise ValidationError(
                message="File specified doesnt exist or is invalid",
                cursor_position=len(document.text),
            )



__all__ = [
    "FileOperationsBase",
    "FileBasedQueryExecutorBase",
    "NotebookExecutorBase",
    "InMemoryQueryExecutorBase",
    "FeatureComputeBase",
    "MetricsBase",
]

from .fileops import *
from .filebased_query_executor import *
from .inmemory_query_executor import *
from .notebook_executor import *
from .feature_compute import *
from .metrics import *



import os
import copy
import json
import numpy as np
from hashlib import sha256
from datetime import datetime
import logging

from enrichsdk import Source

logger = logging.getLogger("app")


class TableSource(Source):
    """
    Load csv/other files into pandas dataframes.

    Parameters specific to this module include:

        * source: A dictionary of dataframe names and how to
          load them. It has a number of attributes:

            * type: Output type. Only 'table' value is
              supported for this option.
            * filename: Output filename. You can use default
              parameters such  runid
            * params: Params are arguments to [pandas read_csv](https://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_csv.html)

    Example::

        ....
        "transforms": {
            "enabled": [
                {
                    "transform": "TableSink",
                    "args": {
                        "source": {
                            "article": {
                                "type": "file",
                                "filename": "%(data)s/ArticleData.csv",
                                "params": {
                                    "delimiter": "|",
                                    "dtype": {
                                        "sku": "category",
                                        "mc_code": "int64",
                                        "sub_class": "category",
                                        "priority": "float64"
                                        ...
                                    }
                                }
                            }
                        }
                      ...
                    }
                }
            ...
           ]
         }

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "TableSource"

    def preload_clean_args(self, args):
        """
        Clean when the spec is loaded...
        """

        # Backward compatability
        if "source" not in args:
            args = {"source": args}

        args = super().preload_clean_args(args)

        # Sanity check...
        assert isinstance(args, dict)
        assert "source" in args
        assert isinstance(args["source"], dict)

        for name, detail in args["source"].items():

            # Insert the frame into the args for backward
            # compatability.
            if (
                ("type" in detail)
                and ("frametype" not in detail)
                and (detail["type"] == "table")
            ):
                detail["frametype"] = "pandas"

            if ("frametype" not in detail) or (detail["frametype"] != "pandas"):
                logger.error(
                    "Invalid configuration. Only pandas table source supported by this source transform",
                    extra=self.config.get_extra({"transform": self.name}),
                )
                raise Exception("Invalid configuration")

            if (
                ("filename" not in detail)
                or (not isinstance(detail["filename"], str))
                or ("params" not in detail)
                or (not isinstance(detail["params"], dict))
            ):
                logger.error(
                    "Invalid args. Filename (string) and params (dict) are required",
                    extra=self.config.get_extra({"transform": self.name}),
                )
                raise Exception("Invalid configuration")

            mapping = {
                "str": str,
                "float": np.float64,
                "float64": np.float64,
                "np.float64": np.float64,
                "np.int64": np.int64,
                "int": np.int64,
                "int64": np.int64,
                "datetime": datetime,
                "category": "category",
            }

            # => Materialize the path...
            detail["filename"] = self.config.get_file(detail["filename"])
            detail["root"] = self.config.enrich_data_dir
            params = detail["params"]
            if "dtype" in params:
                for attr in params["dtype"]:
                    if params["dtype"][attr] in mapping:
                        params["dtype"][attr] = mapping[params["dtype"][attr]]
                    else:
                        params["dtype"][attr] = eval(params["dtype"][attr])

        return args

    def validate_args(self, what, state):
        """ """
        assert isinstance(self.args, dict)
        assert "source" in self.args
        for name, detail in self.args["source"].items():
            assert ("frametype" in detail) and (detail["frametype"] == "pandas")
            assert "filename" in detail
            assert "params" in detail

    def sample_inputs(self):

        frame = self.config.get_dataframe("pandas")

        # Check if nothing has been specified
        if "source" not in self.args:
            return []

        args = self.args["source"]

        # Read 50 rows only...
        args = copy.deepcopy(args)
        for name in args:
            if "params" not in args[name]:
                args[name]["params"] = {}
            args[name]["params"]["nrows"] = 50

        # => Load the tables
        try:
            dfstates = frame.read(args, {})
        except:
            logger.exception("Error while loading the inputs")
            return []

        # => Generate the inputs for the caller...
        inputs = []
        for name in args:
            detail = dfstates[name]
            table = [list(frame.columns(detail["df"]))] + frame.tolist(
                frame.head(detail["df"], 4)
            )
            inputs.append(
                {
                    "name": name,
                    "rows": frame.shape(detail["df"])[0],
                    "columns": frame.shape(detail["df"])[1],
                    "params": detail["params"],
                    "table": table,
                }
            )

        return inputs

    def clean(self, state):
        """
        This is meant for subclass to do some additional processing.
        """
        pass

    def process(self, state):
        """
        Load file...
        """
        # Load all the dataframes. This will use the full enrich
        # deployment's beefed up read function.
        framecls = self.config.get_dataframe("pandas")
        source = self.args["source"]

        dfstates = framecls.read(source, {})
        for dfname, dfstate in dfstates.items():

            # => Insert column description
            columns = dfstate["params"]["columns"]
            for c in columns:
                columns[c]["description"] = self.get_column_description(dfname, c)

            params = dfstate["params"]
            if "filename" in params:
                filename = params["filename"]
            elif "filename" in source.get(dfname, {}):
                filename = self.args[dfname]["filename"]
            else:
                filename = "Unknown"

            detail = {
                "df": dfstate["df"],
                "transform": self.name,
                "frametype": "pandas",
                "params": [
                    params,
                    {
                        "type": "lineage",
                        "transform": self.name,
                        "dependencies": [
                            {"type": "file", "nature": "input", "objects": [filename]}
                        ],
                    },
                ],
                "history": [
                    {
                        "transform": self.name,
                        "log": "Read data using {}".format(framecls.__class__.__name__),
                    }
                ],
            }
            try:
                state.update_frame(dfname, detail, create=True)
            except:
                logger.exception(
                    "Unable to store state",
                    extra=self.config.get_extra({"transform": self.name}),
                )
                raise Exception("Error while loading")

        # Clean the uploaded data...
        self.clean(state)

    def validate_results(self, what, state):
        pass


provider = TableSource



"""
Standard transforms that can be directly included in 
any pipeline.
"""

__all__ = [
    "FileOperations",
    "JSONSink",
    "JSONSource",
    "PQExport",
    "SQLExport",
    "TableSink",
    "TableSource",
]

from .fileops import *
from .jsonsink import *
from .jsonsource import *
from .pqexport import *
from .sqlexport import *
from .tablesink import *
from .tablesource import *



"""
Set of predefined tasks
"""

import os
import pytz
import gzip
import re
import json
import datetime
import requests
import subprocess
import platform
import logging
from enrichsdk.tasks import Task

logger = logging.getLogger("app")

__all__ = ["BackupCoreTask", "SyncLocalCoreTask"]


class CloudMixin(object):
    def get_params(self):

        # resolve the destination
        now = self.config.now()
        params = {
            "node": platform.node(),
            "date": now.strftime("%Y-%m-%d"),
            "datetime": now.strftime("%Y%m%d-%H%M%S"),
        }
        return params

    def validate_args_cloud(self, what, state):

        dirspecs = self.dirspecs
        args = self.args

        fail = False
        msg = ""

        if "aws" not in args and "gcp" not in args:
            fail = True
            msg += "Both 'aws' or 'gcp' are missing. One is required\n"

        if "aws" in args:
            for v in ["credentials", "bucket"]:
                if v not in args["aws"]:
                    fail = True
                    msg += "Element '{}' is missing in aws config\n".format(v)

            try:
                cred = self.get_credentials(args["aws"]["credentials"])

                if ("access_key" not in cred) or ("secret_key" not in cred):
                    fail = True
                    msg += "Invalid S3 credentials. Missing access_key or secret_key\n"

            except Exception as e:
                fail = True
                msg += str(e) + "\n"

        elif "gcp" in args:
            for v in ["boto", "bucket"]:
                if v not in args["gcp"]:
                    fail = True
                    msg += "Element '{}' is missing in gcp config\n".format(v)

        # => Now check the backup/sync specification
        if (dirspecs not in args) or (not isinstance(args[dirspecs], list)):
            fail = True
            msg += "Element '{}' is missing or invalid (not a list)\n".format(dirspecs)
        else:
            for d in args[dirspecs]:
                if not isinstance(d, dict):
                    fail = True
                    msg += "Specification is not a dict: {}".format(d)
                    continue
                if "src" not in d or "dst" not in d:
                    fail = True
                    msg += "Specification should have src and dst"

        return fail, msg

    def validate_args(self, what, state):
        """
        Validate args.

        Looks like 'aws' and 'backupdirs' specification

        Example::

            "args": {
                 "backupdirs": [
                    {
                        "enable": true,
                        "name": "Logs",
                        "src": "%(enrich_root)s/logs/",
                        "dst": "backup/logs/"
                     },
                                ...
                  ],
                  "aws": {
                      "credentials": "enrich-acme",
                      "bucket": "client-enrich"
                  }
            }

        """
        args = self.args

        # Check
        fail, msg = self.validate_args_cloud(what, state)

        if fail:
            logger.error(
                "Invalid configuration",
                extra=self.config.get_extra({"transform": self.name, "data": msg}),
            )
            raise Exception("Invalid configuration")

    def get_env(self):
        """
        Get environment for the cloud provider
        """

        args = self.args
        env = dict(os.environ)
        if "aws" in args:
            cred = self.get_credentials(args["aws"]["credentials"])
            if cred is not None:
                if ("access_key" in cred) and (isinstance(cred["access_key"], str)):
                    env["AWS_ACCESS_KEY_ID"] = cred["access_key"]
                if ("secret_key" in cred) and (isinstance(cred["secret_key"], str)):
                    env["AWS_SECRET_ACCESS_KEY"] = cred["secret_key"]
        else:
            env["BOTO_CONFIG"] = self.config.get_file(args["gcp"]["boto"])

        env["TZ"] = args.get("timezone", "Asia/Kolkata")
        return env

    def get_command(self, d):
        """
        Get the command that must be run
        """

        # To or from cloud
        direction = self.direction

        # Extra params..
        params = self.get_params()

        if "aws" in self.args:
            prefix = "s3"
            bucket = self.args["aws"]["bucket"]
        elif "gcp" in self.args:
            prefix = "gs"
            bucket = self.args["gcp"]["bucket"]
        else:
            raise Exception("Unsupported source/destination")

        # Resolve the source and destination
        if direction == "tocloud":
            src = self.config.get_file(d["src"])
            dst = d["dst"] % params
            dst = "{}://{}/{}".format(prefix, bucket, dst)
        else:
            dst = os.path.abspath(self.config.get_file(d["dst"]))
            src = d["src"] % params
            src = "{}://{}/{}".format(prefix, bucket, src)

        if "aws" in self.args:

            # aws s3 sync s3://mybucket s3://mybucket2
            if direction == "tocloud" and os.path.isfile(src):
                awscmd = ["aws", "s3", "cp", "--quiet"]
            else:
                awscmd = ["aws", "s3", "sync", "--quiet"]

            if self.config.dryrun:
                awscmd.append("--dryrun")

            awscmd.extend([src, dst])
            return awscmd

        if "gcp" in self.args:

            # gsutil sync gs://mybucket gs://mybucket2
            if direction == "tocloud" and os.path.isfile(src):
                gscmd = ["gsutil", "cp"]
            else:
                gscmd = ["gsutil", "-m", "-q", "rsync", "-r"]

            if self.config.dryrun:
                gscmd.append("-n")

            # Add source and destination
            gscmd.extend([src, dst])

            return gscmd

        raise Exception("Unsupported source/destination")

    # => Now run the command
    def run_sync(self, state):
        """
        Run the backup task
        """

        env = self.get_env()

        dirspecs = self.args[self.dirspecs]
        for d in dirspecs:
            enable = d.get("enable", True)
            if not enable:
                continue

            name = d["name"]

            # => What should I be running?
            cmd = self.get_command(d)

            # Now run the command
            p = subprocess.Popen(
                cmd, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            out, err = p.communicate()
            out = str(out, "utf-8")
            err = str(err, "utf-8")

            msg = """\
Cmd: {}
Error
-----
{}

Output:
-------
{}"""
            msg = msg.format(" ".join(cmd), err, out)

            # Log cloud-specific information
            if "BOTO_CONFIG" in env:
                msg += "Boto Config: {}".format(env["BOTO_CONFIG"])

            logger.debug(
                "Backing up {}".format(name),
                extra=self.config.get_extra({"transform": self.name, "data": msg}),
            )

        logger.debug(
            "Sync'd all dirs",
            extra=self.config.get_extra(
                {
                    "transform": self.name,
                }
            ),
        )

    def run(self, state):
        return self.run_sync(state)


class BackupCoreTask(CloudMixin, Task):
    """
    Backsup code and data as required
    """

    NAME = "BackupCoreTask"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "BackupCoreTask"
        self.dirspecs = "backupdirs"
        self.direction = "tocloud"


class SyncLocalCoreTask(CloudMixin, Task):
    """
    Syncs a local directory with  content of a s3 bucket
    """

    NAME = "SyncLocalCoreTask"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "SyncLocalCoreTask"
        self.dirspecs = "localdirs"
        self.direction = "fromcloud"



from functools import partial
import pandas as pd

from .base import *
from .exceptions import *


class TableColumnsExistExpectations(ExpectationBase):
    """
    Check whether table has required columns

    Configuration has to specify a list of columns::

           {
             'expectation': 'table_columns_exist',
             'params': {
                 'columns': ['alpha', 'beta']
                 }
             }
           }

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "table_columns_exist"
        self.description = "Check whether columns exist (in any order)"

    def validate_args(self, config):
        super().validate_args(config)

        # Add extra
        self.args_validator_attrs_list(config, ["columns"])

    def generate(self, df):
        """
        Not implemented
        """
        return {
            "expectation": self.name,
            "description": self.description,
            "params": {"columns": list(df.columns)},
        }

    def validate(self, df, config):
        """
        Check if the specified columns are present in the dataframe
        """

        result = ExpectationResultBase()
        update = partial(
            result.add_result, expectation=self.name, description=self.description
        )

        if not isinstance(df, pd.DataFrame):
            update(passed=False, extra={"reason": "Not a dataframe"})
            return result

        columns = config["params"]["columns"]
        if isinstance(columns, str):
            columns = [columns]

        missing = [c for c in columns if c not in df.columns]
        if len(missing) > 0:
            update(
                passed=False,
                extra={"reason": "Missing columns: {}".format(",".join(missing))},
            )
            return result

        update(passed=True)
        return result


class TableColumnsPositionExpectations(ExpectationBase):
    """
    Check whether table has the right columns in right positions.

    Configuration has to specify the columns and their corresponding
    positions in the test dataframe::

           {
             'expectation': 'table_columns_exist_with_position',
             'params': {
                 'columns': {
                    'alpha': 0,
                    'beta': 2
                 }
             }
           }

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "table_columns_exist_with_position"
        self.description = "Check whether columns exist (in particular order)"

    def validate_args(self, config):
        super().validate_args(config)

        # Add extra
        self.args_validator_attrs_list(config, ["columns"])

    def generate(self, df):
        """
        Not implementated yet.
        """
        columns = df.columns
        columns = {col: position for col, position in enumerate(columns)}
        return {
            "expectation": self.name,
            "description": self.description,
            "params": {"columns": columns},
        }

    def validate(self, df, config):
        """
        Validate the names and positions of columns
        """

        result = ExpectationResultBase()
        update = partial(
            result.add_result, expectation=self.name, description=self.description
        )

        if not isinstance(df, pd.DataFrame):
            update(passed=False, extra={"reason": "Not a dataframe"})
            return result

        columns = config["params"]["columns"]
        if not isinstance(columns, dict):
            raise InvalidConfigurationExpectation("Columns should be a dict")

        invalid = []
        existing = list(df.columns)
        for col, position in columns.items():
            if (
                (not isinstance(position, int))
                or (position >= len(existing))
                or (existing[position] != col)
            ):
                invalid.append(col)

        if len(invalid) > 0:
            update(
                passed=False,
                extra={
                    "reason": "Missing/Wrong position columns: {}".format(
                        ",".join(invalid)
                    )
                },
            )
            return result

        update(passed=True)
        return result



import os
import logging
from jinja2 import Environment, FileSystemLoader, meta

logger = logging.getLogger("app")

# Template rendering functions
def template_render(widgetdir, widgetname, context):
    """
    Render a specified template using the context

    Args:
      widgetname (str): Name of the template
      context (dict): Key-value pairs

    Returns:
      Rendered html template that can be embedded

    """
    filename = widgetname + ".html"
    env = Environment(loader=FileSystemLoader(widgetdir))
    try:
        template = env.get_template(filename)
        return template.render(context)
    except:
        logger.exception("Could not find template")
        return ""


def template_get_variables(widgetdir, widgetname):
    """
    Get the variables from a template

    Args:
       widgetname (str): Name of the template

    Returns:
       List of variables
    """
    filename = widgetname + ".html"
    env = Environment(loader=FileSystemLoader(widgetdir))
    try:
        template_source = env.loader.get_source(env, filename)[0]
        parsed_content = env.parse(template_source)
        return meta.find_undeclared_variables(parsed_content)
    except:
        return []



"""
Get the environment object
"""
import os
import sys
import copy
import yaml
import json


class Context:
    def __init__(self, config=None, *args, **kwargs):

        self.context = {}

        if (config is None) and ("ENRICH_CONTEXT" in os.environ):
            config = os.environ["ENRICH_CONTEXT"]
            if not os.path.exists(config):
                raise Exception("Context file missing: {}".format(config))

        if config is not None:
            if isinstance(config, str):
                try:
                    if config.endswith(".json"):
                        config = json.load(open(config))
                    elif config.endswith(".yaml"):
                        config = yaml.load(open(config), Loader=yaml.FullLoader)
                    else:
                        raise Exception("Unknown input format")
                except:
                    raise
                self.context = config
            elif isinstance(config, dict):
                self.context = config
            else:
                raise Exception("Invalid context file")

        # => Priority yaml config -> environment variables..
        # Root has to be specified
        if "ENRICH_ROOT" not in self.context:
            if "ENRICH_ROOT" not in os.environ:
                raise Exception(
                    "ENRICH_ROOT must be specified in context file or environment"
                )
            else:
                self.context["ENRICH_ROOT"] = os.environ["ENRICH_ROOT"]

        defaults = {
            "ENRICH_RUN_ROOT": "%(ENRICH_ROOT)s",
            "ENRICH_DATA": "%(ENRICH_ROOT)s/data",
            "ENRICH_ETC": "%(ENRICH_ROOT)s/etc",
            "ENRICH_SHARED": "%(ENRICH_ROOT)s/shared",
            "ENRICH_VAR": "%(ENRICH_ROOT)s/var",
            "ENRICH_LIB": "%(ENRICH_ROOT)s/lib",
            "ENRICH_OPT": "%(ENRICH_ROOT)s/opt",
            "ENRICH_LOGS": "%(ENRICH_ROOT)s/logs",
            "ENRICH_CUSTOMERS": "%(ENRICH_ROOT)s/customers",
            "ENRICH_RELEASES": "%(ENRICH_ROOT)s/releases",
            "ENRICH_TEST": "%(ENRICH_ROOT)s/test",
            "ENRICH_MANAGE": "%(ENRICH_ROOT)s/manage",
            "versionmap": "%(ENRICH_ROOT)s/etc/versionmap.json",
            "EMAIL_HOST": "unknown",
            "EMAIL_HOST_USER": "notset",
            "EMAIL_HOST_PASSWORD": "notset",
            "EMAIL_PORT": "9999",
            "EMAIL_USE_TLS": "TRUE",
        }

        for k, v in defaults.items():
            if k not in self.context:
                # If specified in the environment
                if k in os.environ:
                    self.context[k] = os.environ[k]
                else:
                    self.context[k] = v

        # For siteconf check the available alternatives. If none
        # present, then dont add the entry.
        k = "siteconf"
        if k not in self.context:
            if k in os.environ:
                self.context[k] = os.environ[k]
            else:
                for s in [
                    "%(ENRICH_ROOT)s/etc/siteconf.jsonx",
                    "%(ENRICH_ROOT)s/etc/siteconf.json",
                ]:
                    try:
                        s = s % self.context
                        if os.path.exists(s):
                            self.context[k] = s
                            break
                    except:
                        pass

        # Resolve all of them...
        for k, v in self.context.items():
            if isinstance(v, str):
                v = v % self.context
                self.context[k] = v

        if "ENRICH_ROOT" not in self.context:
            print("Please define ENRICH_ROOT or provide a environmental context file")
            raise Exception("Missing ENRICH_ROOT. Cant build context")

    def set_var(self, k, v):
        """
        Update the value for given context variable
        """
        self.context[k] = v

    def get_var(self, k):
        """
        Get the value for given context variable
        """
        return self.context.get(k, None)

    def validate(self):

        for attr in ["ENRICH_ROOT", "ENRICH_DATA", "ENRICH_CUSTOMERS", "ENRICH_ETC"]:
            if attr not in self.context:
                raise Exception("Incomplete context")

    def set_env(self):
        """
        Set the environment using the context
        """
        update = {k: v for k, v in self.context.items() if k.isupper()}
        os.environ.update(update)

    def asdict(self):
        return copy.copy(self.context)



"""
Realtime Feature Engineering

One key objective is to share the same code base with the batch
processing code.

.. warning::
   Module under active development.

"""

from .transforms import *
from .workflow import *
from .db import *
from .log import *
from .messaging import *



"""
Module to facilitate access to Enrich Server API
and also show on the command line.

Visit <server>/api/v1.0/schema to understand the available
API calls.

"""

import json
import requests
import re
from collections import OrderedDict

from .draw import *


class Backend:
    """
    Class to access Enrich Server API
    """

    def __init__(self, config):
        """
        Initialize an access object

        Args:

           config (dict): Should have 'server' and 'key'.
        """

        self.config = config
        assert "server" in config
        assert "key" in config

    def get_api_url(self, path, use_prefix=True):

        path = path.replace("//", "/")
        server = self.config["server"]

        # Avoid double lines...
        if server[-1] == "/" and path[0] == "/":
            path = path[1:]

        if use_prefix:
            apiversion = self.config.get("version", "v1.0")
            url = "{}/api/{}/{}".format(server, apiversion, path)
        else:
            url = "{}/{}".format(server, path)

        return url

    def call(self, url, params={}, method="get", data={}):
        """
        Call backend with some extra query parameters

        Args:

           path (str): API path suffix (after api/v1.0/)
           params (dict): Query parameters
        """

        method = str(method).lower()

        key = self.config["key"]
        headers = {
            "Authorization": "Token {}".format(key),
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

        try:
            if method == "get":
                response = requests.get(
                    url,
                    verify=False,  # ignore SSL checks
                    headers=headers,
                    params=params,
                )
            elif method == "post":
                response = requests.post(
                    url, verify=False, headers=headers, params=params, json=data
                )
            else:
                raise Exception("Only get/post supported")

            if response.status_code not in [200, 400, 404]:
                return {
                    "status": "error",
                    "message": response.content[:100].decode("utf-8"),
                }
            return response.json(object_pairs_hook=OrderedDict)

        except requests.exceptions.ConnectionError as e:
            return {"status": "error", "message": "Connection refused"}
        except Exception as e:
            return {"status": "error", "message": str(e)}

    def status(self, offset=0, page=20):
        """
        Shows a list of tasks/pipelines that ran/are running

        Args:
           offset (int): Offset in the execution list
           page (int): Page in the execution list
        """

        url = self.get_api_url("status")
        result = self.call(url, {"offset": offset, "page": page})

        return result

    def usecases(self):
        """
        Show all the usecases available on server
        """
        url = self.get_api_url("usecases")
        result = self.call(url)
        return result

    def health(self):
        url = self.get_api_url("system/health")
        result = self.call(url)
        return result

    def pipelines(self, usecase):
        """
        Shows all pipelines within a given usecase

        Args:
           usecase (str): Name of the usecase
        """
        url = self.get_api_url("usecase/pipelines")
        result = self.call(url, {"usecase": usecase})
        return result

    def run_list(self, usecase, pipeline):
        """
        Shows all runs of a given pipeline

        Args:
           usecase (str): Name of the usecase
           pipeline (str): Name of the pipeline

        """
        url = self.get_api_url("usecase/pipeline/runs")
        result = self.call(url, {"usecase": usecase, "pipeline": pipeline})
        return result

    def run_detail(self, usecase, pipeline, runid):
        """
        Shows details of a given run of a pipeline

        Args:
           usecase (str): Name of the usecase
           pipeline (str): Name of the pipeline
           runid (str): Name of the run

        """
        url = self.get_api_url("usecase/pipeline/run/detail")
        result = self.call(
            url, {"usecase": usecase, "pipeline": pipeline, "runid": runid}
        )
        return result

    def tasks(self, usecase):
        """
        Shows all tasks within a given usecase

        Args:
           usecase (str): Name of the usecase
        """
        url = self.get_api_url("usecase/tasks")
        result = self.call(url, {"usecase": usecase})
        return result

    def task_run_list(self, usecase, task):
        """
        Shows all runs of a given task

        Args:
           usecase (str): Name of the usecase
           task (str): Name of the task
        """
        url = self.get_api_url("usecase/task/runs")
        result = self.call(url, {"usecase": usecase, "task": task})
        return result

    def task_run_detail(self, usecase, task, runid):
        """
        Shows details of a run

        Args:
           usecase (str): Name of the usecase
           task (str): Name of the task
           runid (str): Name of a run
        """
        url = self.get_api_url("usecase/task/run/detail")
        result = self.call(url, {"usecase": usecase, "task": task, "runid": runid})
        return result



import os
import re
import json
from ..lib import renderlib


class Widgets(object):
    """
    List of available widgets
    """

    pass


############################################
# Update the widget class with attributes.
############################################
thisdir = os.path.dirname(__file__)
templatedir = os.path.join(thisdir, "..", "templates")
templatedir = os.path.abspath(templatedir)
widgetdir = os.path.join(templatedir, "widgets")


def get_func(attrname, filename):
    def dummyfunc():
        pass

    content = open(os.path.join(widgetdir, filename)).read()
    try:
        content = re.search(r"<!--(.*)-->", content).group(1)
    except:
        content = "Widget"

    # Get variables...
    variables = renderlib.template_get_variables(widgetdir, attrname)

    if len(variables) > 0:
        content += """

    Args:

"""
        for v in variables:
            content += "        " + v + "(str): parameter\n"

    content += """
    Returns:

       Rendered html 
"""
    dummyfunc.__doc__ = content
    return dummyfunc


for filename in os.listdir(widgetdir):
    if not filename.endswith(".html"):
        continue

    attrname = filename.replace(".html", "")
    attr = get_func(attrname, filename)
    setattr(Widgets, attrname, property(attr))



import os
import sys
from enrichsdk.policy import *


class PipelineRunResource(BaseResource):
    """
    Resource that implements access to pipeline output
    """

    def __init__(self, run, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.run = run

    def name(self):
        return "PipelineRun::{}::{}".format(self.run.config.name, self.run.name)


class SimpleFilterEngine(BasePolicyEngine):
    """
    This applies the policy on a collection of dataframes
    """

    NAME = "SimpleFilterPolicy"

    def __init__(self, *args, **kwargs):
        self.config = kwargs.pop("config")
        super().__init__(*args, **kwargs)
        self.name = "SingleApplicationPolicy"
        self.filter_handlers = {
            "by_column_value": self.filter_by_column_value,
            "by_column_range": self.filter_by_column_range,
        }

    def filter_by_column_value(self, principal, resource, params):
        """
        Filter the data by value of a particular column. The
        column and the value is specified in policy

        Args:
          principal (obj): Principal object
          resource (obj): Resource object
          params (dict): Filter parameters
        """
        pass

    def filter_by_column_range(self, principal, resource, params):
        """
        Filter the data by the range of values of a particular
        column. The column and the value is specified in policy.

        Args:
          principal (obj): Principal object
          resource (obj): Resource object
          params (dict): Filter parameters
        """
        pass



import os
import json
import argparse
from pyspark.sql import SQLContext

from .spark import start_spark

"""
This is a wrapper around start_spark function. This 
will take the configuration parameters specified in 
'config' and also enrich environment to create the 
spark session 
"""


def start_spark_session(config):
    """
    Main script

    :return: None
    """

    # Collect parameters
    name = config["name"]

    if "master" in config:
        master = config["master"]
    elif "ENRICH_SPARK_MASTER" in os.environ:
        master = os.environ["ENRICH_SPARK_MASTER"]
    else:
        master = "local"

    jar_packages = config.get("jar_packages", [])
    spark_config = config.get("spark_config", {})

    # start Spark application and get Spark session, logger and config
    spark, log, extra_config = start_spark(
        app_name=name,
        master=master,
        jar_packages=jar_packages,
        spark_config=spark_config,
    )

    # Save it...
    sc = spark.sparkContext
    sqlsc = SQLContext(sc)

    config["spark"] = spark
    config["sc"] = sc
    config["sqlsc"] = sqlsc
    config["applicationid"] = sc.applicationId
    config["logger"] = log

    # log that main ETL job is starting
    log.warn("Pipeline is up-and-running")

    return


def stop_spark_session(config):

    spark = config["spark"]
    log = config["log"]

    log.warn("test_etl_job is finished")

    spark.stop()

    return None



import os
import sys
import json
import re
from datetime import datetime
import traceback
import pandas as pd
from s3fs import S3FileSystem
from ..lib.exceptions import NoDataFound

from dateutil import parser as dateparser
from datetime import datetime, date, timedelta


class DataSource(object):
    """
    Class for specifying a dataset. This is a base class
    meant to be derived and implemented.
    """

    def __init__(self, params, *args, **kwargs):

        self._type = params.get("type", "unknown")
        """
        Type of the dataset.
        """

        self.name = params["name"]
        """
        Unique name of the dataset, e.g., Events
        """

        self.description = params.get("description", "")
        """
        str: Description
        Example: "Daily events for Alpha Sensor"
        """

        alt_names = params.get("alt_names", [])
        if not isinstance(alt_names, list):
            alt_names = [alt_names]
        alt_names = [str(x) for x in alt_names]
        self.alt_names = alt_names
        """
        list: Alternative names for the dataset
        Example: "['txn_charges']"
        """

        self.version = params.get("version", "v1")
        """
        str: Version
        Example: "v1"
        """

        self.resolve = kwargs.get("resolve", params.get("resolve", {}))
        """
        Parameters to resolve paths
        """

        # Backward compatability
        if "paths" not in params:
            params["paths"] = []
            for name in ["local", "test", "backup"]:
                if name in params:
                    params["paths"].append(
                        {
                            "name": name,
                            "nature": "s3" if name == "backup" else "local",
                            "path": params[name],
                        }
                    )

        self.validate(params)

        self._type = "File"

        self.isfile = params.get("isfile", False)
        """
        Specifies whether the dataset is simple (a file for each
        run) or more complex directory hierarchy.

        Example::

            # A complex dataset
            Dataset(params={
                "name": 'athena-availability',
                "isfile": False,
                ...

        """
        self.issortable = params.get("issortable", True)

        self.subsets = params.get("subsets", [])
        """
        A dataset oftentimes has multiple components. Define
        them here. We can give a name to each.

        Example::

          {
             ...
             "subsets": [
                 {
                     "name": "v1culltable",
                     "pattern": ".*v1culltable.csv$"
                 },...
              ]
          }
        """

        self.paths = params.get("paths", {})
        """
        dict: Paths to be resolved
        Example::

            [
              {
                "name": "test",
                "nature": "local",
                "path": "%(data_root)s/shared/datasets/athena/v2/availability"
              }
           ]
         """

        self.registry = None
        """
        object: Dataset manager object
        """

    def matches(self, names):
        """
        Check if the dataset has a particular name
        """
        if not isinstance(names, list):
            names = [names]

        matches = [((name == self.name) or (name in self.alt_names)) \
                                   for name in names]
        return any(matches)

    def set_resolve(self, resolve):
        """
        Allow the user to specify a resolving dict
        """
        assert isinstance(resolve, dict)
        assert len(resolve) > 0
        self.resolve.update(resolve)

    def set_registry(self, registry):
        """
        Set the registry object
        """
        self.registry = registry

    def get_paths(self):
        """
        Return the available paths (full list of dictionaries,
        not just the names
        """
        return self.paths

    def get_path_by_name(self, name, full=False, resolve=None):
        """
        Return the path definition for a path with a given name
        """

        if resolve is None:
            resolve = self.resolve

        selected = None
        for p in self.paths:
            if p["name"] == name:
                selected = p
                break
        if selected is None:
            raise Exception("Path with name {} not present".format(name))

        if full:
            return selected

        path = selected["path"]

        try:
            if (
                (resolve is not None)
                and (isinstance(resolve, dict))
                and (len(resolve) > 0)
            ):
                path = path % resolve
        except:
            raise Exception("Path could not be resolved")

        return path

    def get_path_by_nature(self, nature, full=True, resolve=None):

        if resolve is None:
            resolve = self.resolve

        selected = None
        for p in self.paths:
            if p["nature"] == nature:
                selected = p
                break

        if selected is None:
            raise Exception("Path with nature {} not present".format(nature))

        if full:
            return selected

        path = selected["path"]

        try:
            if (
                (resolve is not None)
                and (isinstance(resolve, dict))
                and (len(resolve) > 0)
            ):
                path = path % resolve
        except:
            raise Exception("Path could not be resolved")

        return path

    def has_subsets(self):
        """
        Does this dataset have components?
        """
        return len(self.subsets) > 0

    def get_subsets(self):
        """
        Return the names of subsets available
        """
        return [s["name"] for s in self.subsets]

    def get_subset_description(self, name):
        """
        Get the description for a subset
        """
        for s in self.subsets:
            if s["name"] == name:
                return s.get("description", self.description)

        raise Exception("Unknown subset: {}".format(name))

    def in_subset(self, name, spec):
        """
        Return the names of subsets available
        """

        if (not isinstance(spec, dict)) or (len(spec) == 0) or ("filename" not in spec):
            raise Exception("invalid subset check input. Expecting dict with filename")

        # Does the subset name exist?
        subset = None
        for s in self.subsets:
            if name == s["name"]:
                subset = s
                break
        if subset is None:
            return False

        #
        if "pattern" not in subset:
            return False

        filename = spec["filename"]
        pattern = subset["pattern"]
        return re.search(pattern, filename) is not None

    def get_path(self, name, full=False, resolve=None):
        return self.get_path_by_name(name, full, resolve)

    def validate(self, params):
        """ """
        if not isinstance(params, dict):
            raise Exception("Invalid parameters for dataset")

        for col in ["name", "paths"]:
            if col not in params:
                raise Exception("Data specification should have {}".format(col))

        if not isinstance(params["paths"], list):
            raise Exception("Paths should be a list of dictionaries")

        for elem in params["paths"]:
            if not isinstance(elem, dict) or len(elem) == 0:
                raise Exception("paths element should be non-empty dict")
            if ("nature" not in elem) or (elem["nature"] not in self.valid_natures):
                raise Exception(
                    f"paths element have nature {'|'.join(self.valid_natures)}"
                )
            if "path" not in elem:
                raise Exception("paths element should have a 'path'")

    def __str__(self):
        return "[{}] {}".format(self._type, self.name)


class Dataset(DataSource):
    """
    Class for specifying a dataset. This is typically
    an input to the transforms.

    Usage::

        # A dataset that has one directory for each day. Within that
        # there are atleast two sub-datasets
        Dataset(params={
            "name": 'athena-availability',
            "type": "file",
            "paths": [
                {
                    "name": "test",
                    "nature": "local",
                    "path": "%(data_root)s/shared/datasets/athena/v2/availability",

                },
                {
                    "name": "local",
                    "nature": "local",
                    "path": "%(enrich_data_dir)s/acme/Marketing/shared/datasets/athena/v2/availability",
                },
                {
                    "name": "s3",
                    "nature": "s3",
                    "path": "%(backup_root)s/%(node)s/data/acme/Marketing/shared/datasets/athena/v2/availability",
                },
            ],
            "match": {
                "generate": "generate_datetime_daily",
                "compare": "compare_datetime_pattern_match_range",
                "pattern": "%Y-%m-%d",
            },
            "subsets": [
                {
                    "name": "CatalogScore",
                    "pattern": ".*catalogscore.csv$",
                    "description": "Score of all products in the catalog"
                },
                {
                    "name": "ProductWeight",
                    "pattern": ".*productweight.csv$",
                    "description": "Assortment weight for each product"
                }
            ]
        })

    """

    def __init__(self, params, *args, **kwargs):

        super().__init__(params, *args, **kwargs)

        self.match = params.get("match", {})
        """
        dict: Generating and matching rules

        Example::

            {
                "generate": "generate_datetime_daily",
                "match": "match_datetime_pattern_range",
                "pattern": "plpevents-%Y%m%d-%H%M%S",
            }
        """

    def sample(self, filename, safe=True, fd=None, nrows=10):
        """
        Sample a file belonging to this dataset.Subclass and
        overload this function if the data is sensitive.

        Args:
            filename (str): A file that belongs to this dataset
            safe (bool): Whether file is trusted
            fd (object): File descriptor for s3/gcs/other non-filesystems
            nrows (int): Number of rows to sample
        """

        if fd is None:
            # Text formats..
            if filename.split(".")[-1].lower() in ["csv", "tsv"]:
                fd = open(filename)
            else:
                fd = open(filename, "rb")

        # Read the file
        if filename.lower().endswith(".csv"):
            df = pd.read_csv(fd, nrows=nrows)
        elif filename.lower().endswith(".csv.gz"):
            df = pd.read_csv(fd, nrows=nrows, compression="gzip", error_bad_lines=False)
        elif filename.lower().endswith(".tsv.gz"):
            df = pd.read_csv(
                fd, nrows=nrows, sep="\t", compression="gzip", error_bad_lines=False
            )
        elif filename.lower().endswith(".tsv"):
            df = pd.read_csv(fd, nrows=nrows, sep="\t")
        elif filename.lower().endswith(".pq") or filename.lower().endswith(
            ".pq.sample"
        ):
            df = pd.read_parquet(fd)
            if nrows is not None:
                df = df.head(nrows)
        else:
            raise Exception(f"Unknown file format: {filename}")

        return df

    @property
    def local(self):
        """
        Read the path with the name 'local'. Available for
        backward compatability
        """
        return self.get_path("local")

    @property
    def test(self):
        """
        Read the path with the name 'test'. Available for
        backward compatability
        """
        return self.get_path("test")

    @property
    def backup(self):
        """
        Read the path with the name 'backup'. Available for
        backward compatability
        """
        return self.get_path("backup")

    @property
    def valid_natures(self):
        return ["local", "remote", "s3", "gcs"]

    def validate(self, params):
        """
        Validate dataset arguments. It should be a dictionary
        with atleast three elements: name, match, and paths.

        Args:
          params (dict): Parameters for dataset

        Params dict should have the following:

          * name: string
          * match: dictionary with generate function (name or lambda), are function (name or lambda), and pattern (string).

        """

        super().validate(params)

        # If match is specified, then do something...
        if ("match" not in params) or (not isinstance(params["match"], dict)):
            raise Exception("Match should be a valid dictionary")

        if "pattern" not in params["match"]:
            raise Exception("Matching requirement match.pattern missing")

        pattern = params["match"]["pattern"]
        if not isinstance(pattern, str) and not callable(pattern):
            raise Exception("Invalid pattern specification")

        generate = params["match"].get("generate", "generate_datetime_daily")

        if ((isinstance(generate, str)) and (not hasattr(self, generate))) and (
            not callable(generate)
        ):
            raise Exception("Matching requirement {} missing/invalid".format(generate))

        compare = params["match"].get("compare", "compare_datetime_pattern_match_range")
        if ((isinstance(compare, str)) and (not hasattr(self, compare))) and (
            not callable(compare)
        ):
            raise Exception("Matching requirement {} missing/invalid".format(compare))

    def generate(self, start, end=None, full=False, name="default", resolve=None):
        """
        Given a start and an end, generate the datetime
        objects corresponding to each run

        Args:
           start (datetime): Start of the time range
           end (datetime): End of range. If None, default = start
           full (bool): Whether full path is required or only the suffix. Default is false
           name (str): Name of the path specification. Optional is full path is required
           resolve (dict): Additional path resolution parameters

        Returns:
           list (dict): List of dictionaries (name, timestamp)

        """
        if (not isinstance(start, (datetime, date))) or (
            (end is not None) and (not isinstance(end, (datetime, date)))
        ):
            raise Exception("Start or end is not a datetime/date")

        # One day
        if end is None:
            end = start

        # Handle ordering...
        if end < start:
            start, end = end, start

        func = self.match["generate"]
        if isinstance(func, str):
            func = getattr(self, func)

        # Generate all possible names
        names = func(start, end)

        # Now insert the full path...
        try:
            if full:
                path_root = self.get_path_by_name(name, resolve=resolve)
                for n in names:
                    n['path'] = os.path.join(path_root, n['name'])
        except:
            logger.exception("Unable to resolve names")

        # Check if the name is in the names..
        return names

    def generate_datetime_daily(self, start, end):
        """
        Given a start and an end, generate the datetime
        objects corresponding to each run for each day.

        Args:
           start (datetime): Start of the time range
           end (datetime): End of range. If None, default = start
        """

        pattern = self.match["pattern"]

        names = []
        current = start
        while current <= end:
            if isinstance(pattern, str):
                name = current.strftime(pattern)
            elif callable(pattern):
                name = pattern(current)
            else:
                raise Exception("Unsupported pattern type")

            names.append({"timestamp": current.isoformat(), "name": name})
            current += timedelta(days=1)

        return names

    def compare_datetime_pattern_match_range(self, name, start, end):
        """
        Given a start and an end, generate the datetime
        objects corresponding to each run for each day, and
        check whether a named file/directory exists in that list.

        Args:
           start (datetime): Start of the time range
           end (datetime): End of range. If None, default = start
        """

        if not isinstance(name, str):
            return False

        # Generate all possible <timestamp>: <name> combinations
        names = self.generate(start, end)

        # names is a dict {'timestamp': '2021-02-02T13:24:11', 'name': '2021-02-02'}
        names = [n["name"] for n in names]
        # Check if the name is in the names or in the range. The
        # latter is required if there are multiple runs in a given day
        #
        return (name in names) or ((name >= min(names) and name <= max(names)))

    def match_content(self, fs, localdir, backupdir, dirname):
        """
        Check whether local path is replicated in the  s3/blob store

        Args:
           fs (object): s3fs handle
           localdir (str): Root local dir to check
           backupdir (str): Root remote dir in block store to check
           dirname (str): path within the localdir
        """

        localpath = os.path.join(localdir, dirname)
        if os.path.isfile(localpath):
            filesize = getsize(localpath)
            backupfile = os.path.join(backupdir, dirname)
            try:
                detail = fs.info(backupfile)
            except FileNotFoundError:
                return False, "Missing file in backup ({})".format(dirname)
            except Exception as e:
                return False, str(e)

            if filesize != detail["size"]:
                return False, "Filesize mismatch: {}".format(dirname)

            return True, "Matched {} (size: {}M)".format(
                dirname, round(filesize / 1000000, 1)
            )

        matched = 0
        totalsize = 0
        for root, dirs, files in os.walk(localpath):
            for name in files:
                filename = os.path.join(root, name)
                relpath = os.path.relpath(filename, start=localpath)
                backupfile = os.path.join(backupdir, dirname, relpath)
                filesize = getsize(filename)
                try:
                    detail = self.fs.info(backupfile)
                except FileNotFoundError:
                    return False, "Missing file in backup ({})".format(relpath)
                except Exception as e:
                    return False, str(e)

                if filesize != detail["size"]:
                    return False, "Filesize mismatch: {}".format(relpath)

                matched += 1
                totalsize += filesize

        return True, "Matched {} (size: {}M)".format(
            matched, round(totalsize / 1000000, 1)
        )

    def compare(self, name, start, end=None):
        """
        Given a start and an end, generate the datetime
        objects corresponding to each run

        Args:
           name (str): Directory name
           start (datetime): Start of the time range
           end (datetime): End of range. If None, default = start

        """
        if (not isinstance(start, datetime)) or (
            (end is not None) and (not isinstance(end, datetime))
        ):
            raise Exception("Generate names in the range")

        # One day
        if end is None:
            end = start

        # Handle ordering...
        if end < start:
            start, end = end, start

        func = params["match"].get("compare", "compare_datetime_pattern_match_range")
        if isinstance(func, str):
            func = getattr(self, func)

        # Generate all possible names
        return func(name, start, end)

    def listdir(self, name, fshandle=None, resolve=None, detail=False):
        """
        List the dataset directory
        """

        selected = None
        for p in self.paths:
            if p["name"] == name:
                selected = p
                break

        if selected is None:
            raise Exception("Unknown name of path: {}".format(name))

        if (selected["nature"] in ["s3", "gcs"]) and (
            (fshandle is None) or (not isinstance(fshandle, S3FileSystem))
        ):
            raise Exception(
                "Filesystem handle should be provided when nature is s3|gcs"
            )

        if not isinstance(detail, bool):
            raise Exception("detail is a boolean field")

        path = self.paths[name]

        try:
            if (
                (resolve is not None)
                and (isinstance(resolve, dict))
                and (len(resolve) > 0)
            ):
                path = path % resolve
        except:
            raise Exception("Path could not be resolved")

        if fshandle is None:
            return os.listdir(path)

        results = fshandle.listdir(path, detail=detail)
        if not detail:
            results = [os.path.basename(r) for r in results]

        return results

    def read_data(
        self,
        start_date,
        end_date,
        filename,
        readfunc,
        errors=True,
        name="default",
        params={},
        resolve={},
    ):
        """
        Read a single dataset

        Args:
             start_date (str): Starting date to scan
             end_date (str): Ending date of the scane
             filename (str): Filename within each day's directory
             readfunc (method): Callback
             errors (bool): What to do on failure. True = bailout
             name (str): Which subset to look at?
             params (dict): Params to the readfunc
             resolve (dict): Parameters for resolution of path

        Returns:
             dataframe: Data read
             dict: Optional metadata on the read

        """

        dates = self.generate(start_date, end_date)
        if len(dates) == 0:
            raise Exception("No dates found")

        # Get the path template...
        path = self.get_path_by_name(name, resolve)

        metadata = {"start": datetime.now(), "files": []}
        # Now compose all the dataframes...
        _dfs = []
        for dt in dates:
            fullpath = os.path.join(path, dt["name"], filename)
            try:
                metadata["files"].append(fullpath)
                df = readfunc(fullpath, params, date=dt["name"])
                if (df is None) or (len(df) == 0):
                    continue
            except:
                if errors:
                    raise
            _dfs.append(df)

        if len(_dfs) == 0:
            raise NoDataFound()

        df = pd.concat(_dfs, ignore_index=True)
        del _dfs

        metadata["end"] = datetime.now()

        # Note the access metadata
        if ((self.registry is not None) and
            (hasattr(self.registry, "access")) and
            (callable(self.registry.access))):
            self.registry.access(self, metadata, nature="read")

        return df, metadata

    def write_data_note(
        self,
            metadata
    ):
        # Note the access metadata
        if ((self.registry is not None) and
            (hasattr(self.registry, "access")) and
            (callable(self.registry.access))):
            self.registry.access(self, metadata, nature="read")


class DatabaseTable(DataSource):
    def __init__(self, params, *args, **kwargs):
        super().__init__(params, *args, **kwargs)
        self._type = "DBTable"

    @property
    def valid_natures(self):
        return ["db"]


class Arg:
    """
    Dataset command argument
    """

    def __init__(self, name, description, *args, **kwargs):
        self.name = name
        self.description = description

    def find(self, datasource):
        """
        Find value values
        """
        raise Exception("Not implemented")

    def validate(self, datasource, value):
        raise Exception("Not implemented")

    def resolve(self, datasource, value, resolve=None):
        raise Exception("Not implemented")


class DateArg(Arg):
    """
    Datetime
    """

    def __init__(self, name, *args, **kwargs):
        super().__init__(name, "Date within past 1 year", *args, **kwargs)

    def find(self, datasource):
        today = datetime.now().date()
        return today

    def validate(self, datasource, value):
        today = datetime.now()
        try:
            dt = dateparser.parse(value)
            diff = today - dt
            if (diff.days > 365) or (diff.days < 0):
                return False
            return True
        except:
            return False

    def resolve(self, datasource, value, resolve):
        dt = dateparser.parse(value)
        return dt


class DatasetPathNatureArg(Arg):
    def __init__(self, name, nature, *args, **kwargs):

        super().__init__(name, "Path of type: {}".format(nature), *args, **kwargs)
        self.nature = nature

    def find(self, datasource):

        valid = []
        for p in datasource.paths:
            if p["nature"] == self.nature:
                valid.append(p["name"])

        return valid

    def validate(self, datasource, value):

        if not isinstance(datasource, Dataset):
            return False

        for p in datasource.paths:
            if (p["name"] == value) and (p["nature"] == self.nature):
                return True

        return False

    def resolve(self, datasource, value, resolve=None):

        if not isinstance(datasource, Dataset):
            raise Exception("Invalid datasource. Expected dataset")

        return datasource.get_path_by_name(value, resolve=resolve)


def get_commands():

    src_s3 = DatasetPathNatureArg("src", "s3")
    dst_s3 = DatasetPathNatureArg("dst", "s3")
    src_local = DatasetPathNatureArg("src", "local")
    dst_local = DatasetPathNatureArg("dst", "local")
    start = DateArg("start")
    end = DateArg("end")

    commands = [
        {
            "source_type": ["File"],
            "args": [src_s3, dst_local, start, end],
            "name": "s3-to-local",
            "description": "Download to local data directory",
            "command": "aws s3 sync s3://%(source)s/%(subdir)s%(slash)s %(destination)s/%(subdir)s%(slash)s",
        },
        {
            "source_type": ["File"],
            "args": [src_local, dst_s3, start, end],
            "name": "local-to-s3",
            "description": "Upload from local data directory to s3",
            "command": "aws s3 sync %(source)s/%(subdir)s%(slash)s s3://%(destination)s/%(subdir)s%(slash)s",
        },
        {
            "source_type": ["File"],
            "args": [src_s3, dst_s3, start, end],
            "name": "s3-to-s3",
            "description": "Sync from one s3 path to another s3 path",
            "command": "aws s3 sync s3://%(source)s/%(subdir)s%(slash)s s3://%(destination)s/%(subdir)s%(slash)s",
        },
        {
            "source_type": ["File"],
            "args": [src_local, dst_local, start, end],
            "name": "remote-to-local",
            "description": "Copy from remote local to test",
            "command": "scp -C -r ubuntu@%(node)s:%(source)s %(destination)s/%(subdir)s",
        },
        {
            "source_type": ["File"],
            "args": [src_local, dst_local, start, end],
            "name": "local-to-remote",
            "description": "Copy from test to remote local path",
            "command": "scp -C -r %(destination)s/%(subdir)s ubuntu@%(node)s:%(source)s",
        },
    ]

    return commands


class DatasetRegistry(object):
    """
    Registry for datasets.

    This provides a search and resolution interface.

    There is a notion of a command - a scp/aws command line template.
    The registry allows enumeration of the commands and enables
    scripting.

    """

    def __init__(self, *args, **kwargs):
        """

        Args:
            commands (list): List of command templates. Optional. If not specified,
                             the system will use defaults.
            resolve (dict): Path resolution dictionary

        """
        self.commands = kwargs.get("commands", get_commands())
        self.resolve = kwargs.get("resolve", {})
        self.state = kwargs.get('state', None)
        self.transform = kwargs.get('transform', None)
        if ((self.state is None) and (hasattr(self.transform, 'state'))):
            self.state = getattr(self.transform, 'state')

        self.datasets = []
        self.validate()

    def validate(self):
        for c in self.commands:
            assert isinstance(c, dict)
            assert ("source_type" in c) and (isinstance(c["source_type"], list))
            assert "args" in c
            assert isinstance(c["args"], list)
            assert "name" in c
            assert "command" in c
            assert isinstance(c["command"], str)
            assert len(c["command"]) > 0

    def get_command(self, name, source_type="File"):
        """
        Get the command specification by specifying its name

        Args:
            name (str): Name of the command template
        """
        for c in self.commands:
            if source_type not in c["source_type"]:
                continue
            if c["name"] == name:
                return c
        raise Exception("Missing command: {}".format(name))

    def get_commands(self, source_type="File"):
        """
        Get the command specifications as a list of dicts

        """
        if source_type is None:
            return self.commands

        # If specified, then apply that filter..
        return [c for c in self.commands if source_type in c["source_type"]]

    def add_datasets(self, items):
        """
        Add to the registry. Each element in the item list
        should be a Dataset (or subclass).

        Args:
           items (list): A list of datasets
        """
        if isinstance(items, list):
            for item in items:
                assert isinstance(item, DataSource)
                item.set_resolve(self.resolve)
                item.set_registry(self)
            self.datasets.extend(items)
        else:
            assert isinstance(items, DataSource)
            items.set_resolve(self.resolve)
            item.set_registry(self)
            self.datasets.append(items)

    def set_resolve(self, resolve):
        """
        Set the resolution parameters

        Args:
           resolve (dict): Resolution parameters

        """
        assert isinstance(resolve, dict)
        assert len(resolve) > 0
        self.resolve = resolve

    @property
    def params(self):
        return self.resolve

    def set_params(self, params):
        """
        Deprecated. Use set_resolve
        """
        return self.set_resolve(params)

    def list(self):
        """
        List all datasets in the registry
        """
        return [str(d) for d in self.datasets]

    def find(self, names):
        """
        Find a dataset in the registry.

        Args:
            names (str): Name of the dataset
        """
        if not isinstance(names, (list, str)):
            raise Exception(f"Unexpected dataset name(s): {names}")

        for d in self.datasets:
            if d.matches(names):
                return d

        return None

    def access(self, dataset, metadata, nature='read-write'):

        if self.state is None:
            # Nothing to be done
            return

        self.state.make_dataset_access_note(dataset, metadata, nature)



"""
Dummy task 
"""

import time
import logging
from enrichsdk.tasks import Task

logger = logging.getLogger("app")

__all__ = ["DummyTask"]


class DummyTask(Task):
    """
    Backsup code and data as required
    """

    NAME = "DummyTask"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "Dummy"
        self.testdata = {
            "data": {},
            "conf": {"enable_extra_args": False, "args": {"sleep": 10}},
        }

        self.supported_extra_args = [
            {
                "name": "jobid",
                "description": "Export JobID",
                "type": "str",
                "required": True,
                "default": "22832",
            }
        ]

    def validate_args(self, what, state):
        """
        Validate args.

        Example::

            "args": {
                 "sleep": 10
            }

        """
        args = self.args
        fail = False
        msg = ""
        if "sleep" not in args:
            fail = True
            msg += "Argument passed does not have/is invalid sleep value\n"

        if fail:
            logger.error(
                "Invalid params",
                extra=self.config.get_extra({"transform": self.name, "data": msg}),
            )
            raise Exception("Invalid params")

    def run(self, state):
        """
        Run the backup task
        """

        # Get the number of seconds to sleep for
        sleep_val = self.args["sleep"]

        # Check if the sleep argument is an integer
        assert isinstance(sleep_val, int)

        time.sleep(sleep_val)

        logger.debug(
            "Sleep task complete.",
            extra=self.config.get_extra(
                {
                    "transform": self.name,
                }
            ),
        )



import os
import sys
import json
import yaml
import copy
import tempfile
import shutil
import time
import glob
import re
import traceback
import subprocess
import numpy as np
import pandas as pd
from enrichsdk import Compute, S3Mixin
from datetime import datetime, date, timedelta
from dateutil import parser as dateparser, relativedelta
import logging
from sqlalchemy import create_engine, text as satext
from functools import partial

logger = logging.getLogger("app")

from enrichsdk.utils import get_lineage_of_query
from enrichsdk.lib.exceptions import NoDataFound


def note(df, title):
    msg = title + "\n"
    msg += "--------" + "\n"
    msg += "Timestamp: " + str(datetime.now()) + "\n"
    msg += "\nShape: " + str(df.shape) + "\n"
    msg += "\nColumns: " + ", ".join(df.columns) + "\n"
    if len(df) > 0:
        msg += "\nSample:" + "\n"
        msg += df.sample(min(2, len(df))).T.to_string() + "\n" + "\n"
    msg += "\nDtypes" + "\n"
    msg += df.dtypes.to_string() + "\n"
    msg += "------" + "\n"
    return msg


class CustomEncoder(json.JSONEncoder):
    def default(self, obj):
        try:
            return super().default(obj)
        except:
            return str(obj)


def get_yesterday():
    yesterday = date.today() + timedelta(days=-1)
    return yesterday.isoformat()


def get_today():
    return date.today().isoformat()


class MetricsBase(Compute):
    """
    Compute metrics as input for the anomaly/other computation

    Features of transform baseclass include:

        * Flexible configuration
        * Highlevel specification of dimensions and metrics

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "MetricsBase"
        self.description = "Compute metrics against datasources"
        self.testdata = {
            "data_root": os.path.join(os.environ["ENRICH_TEST"], self.name),
            "statedir": os.path.join(os.environ["ENRICH_TEST"], self.name, "state"),
            "conf": {"args": {}},
            "data": {},
        }

    @classmethod
    def instantiable(cls):
        return False

    def get_printable_db_uri(self, engine):
        """
        pretty print the URL
        """
        username = engine.url.username
        host = engine.url.host
        database = engine.url.database
        drivername = engine.url.get_driver_name()

        return f"{drivername}:///{host}/{database}/"

    def get_db_uri(self, source):
        """
        Return database URI for a source
        """
        return source["uri"]

    def get_handlers(self, profile):
        """
        Define various callbacks that take a dataframe, spec
        and compute. Specific to a single profile.
        """
        return {}

    def get_profile(self):
        """
        Read the profile json
        """

        if (not hasattr(self, "profiledir")) and (not hasattr(self, "profilefile")):
            raise Exception(
                "'profiledir' transform attribute should be defined to use default get_profile method"
            )

        paths = []
        if hasattr(self, "profilefile"):
            paths.append(self.profilefile)

        if hasattr(self, "profiledir"):
            paths.extend(
                [
                    self.profiledir + "/profile.json",
                    self.profiledir + "/profile.yaml",
                    self.profiledir + "/profilespec.json",
                    self.profiledir + "/profilespec.yaml",
                ]
            )

        profile = None
        for p in paths:
            if not os.path.exists(p):
                continue
            if p.endswith(".json"):
                profile = json.load(open(p))
            elif p.endswith(".yaml"):
                profile = yaml.load(open(p))

        if profile is None:
            raise Exception("Profile could not be found")

        sources = profile["sources"]
        specs = profile["specs"]
        logger.debug(
            f"Found {len(sources)} sources {len(specs)} specs",
            extra={"transform": self.name, "data": json.dumps(profile, indent=4)},
        )

        return profile

    def get_specs(self, profile):
        if (not isinstance(profile, dict)) or ("specs" not in profile):
            raise Exception("Specs not defined in profile")
        return profile["specs"]

    def get_sources(self, profile):
        if (not isinstance(profile, dict)) or ("sources" not in profile):
            raise Exception("Sources not defined in profile")
        return profile["sources"]

    def get_db_query(self, source):

        # Generate/extract the query...
        query = source["query"]
        if callable(query):
            query = quer(source)
        return query

    def update_frame(self, source, df, lineage=None):

        if isinstance(source, str):
            name = source
            description = ""
        else:
            name = source["name"]
            description = source.get("description", "")

        if self.state.has_frame(name):
            return

        params = self.get_column_params(name, df)
        if lineage is not None:
            if isinstance(lineage, dict):
                params.append(lineage)
            else:
                params.extend(lineage)

        detail = {
            "name": name,
            "df": df,
            "frametype": "pandas",
            "description": description,
            "params": params,
            "transform": self.name,
            "history": [],
        }

        self.state.update_frame(name, detail)

    def read_db_source(self, source):

        # Get the SQLAlchemy URI
        uri = self.get_db_uri(source)

        # Get the query
        query = self.get_db_query(source)

        # Create the engine
        engine = create_engine(uri)

        # Run the query
        df = pd.read_sql_query(satext(query), con=engine)

        # Now the input load...
        lineage = {
            "type": "lineage",
            "transform": self.name,
            "dependencies": [
                {
                    "type": "database",
                    "nature": "input",
                    "objects": [self.get_printable_db_uri(uri)],
                },
            ],
        }

        self.update_frame(source, df, lineage)

        return df

    def get_datasets(self, profile, specs):
        """
        Load the datasets specified by the profile
        """

        if not isinstance(profile, dict) or len(profile) == 0:
            logger.warning("Empty profile", extra={"transform": self.name})
            return {}

        # Get various kinds of handlers..
        handlers = self.get_handlers(profile)
        if not isinstance(handlers, dict) or len(handlers) == 0:
            logger.warning("No handlers specified", extra={"transform": self.name})
            handlers = {}

        required_sources = []
        for s in specs:
            required_sources.extend(s["sources"])
        required_sources = list(set(required_sources))

        # Now no about constructucting the datasets
        datasets = {}

        sources = self.get_sources(profile)
        for source in sources:

            nature = source.get("nature", "db")
            name = source["name"]

            if name not in required_sources:
                continue

            pipeline = source.get("pipeline", None)
            generate = source.get("generate", None)

            # Only db is used for now...
            try:
                if nature == "db":
                    result = self.read_db_source(source)
                elif (
                    (generate is not None)
                    and (generate in handlers)
                    and (callable(handlers[generate]))
                ):
                    result = handlers[generate](source)
                elif (generate is not None) and (hasattr(self, generate)):
                    result = getattr(self, generate)(source)
                else:
                    raise Exception(f"Invalid specification: {name}")
            except:
                logger.exception(
                    f"[{name}] generation failed", extra={"transform": self.name}
                )
                continue

            # Clean the read the dataset...
            try:
                if pipeline is not None and isinstance(pipeline, list):
                    for processor in pipeline:
                        if isinstance(processor, str):
                            if processor in handlers:
                                result = handlers[processor](result, source)
                            elif hasattr(self, processor):
                                result = getattr(self, processor)(result, source)
                            else:
                                raise Exception(f"Missing post-processor: {processor}")
                        elif callable(processor):
                            result = processor(result, source)
                        else:
                            raise Exception(
                                "Only method names/callables are supported are supported"
                            )
            except:
                logger.exception(
                    f"[{name}] post-processing failed", extra={"transform": self.name}
                )
                continue

            # We could return multiple values or a single value
            if isinstance(result, dict):
                datasets.update(result)
            else:
                datasets[name] = df

        return datasets

    def get_spec_sources(self, spec, datasets):

        name = spec["name"]

        if ("sources" not in spec) and ("source" not in spec):
            raise Exception(f"[{name}] Invalid specification. Missing dataset")

        sources = spec.get("sources", spec.get("source"))
        if isinstance(sources, str):
            sources = [sources]

        policy = spec.get("missing", "fail")
        for s in sources:
            if s not in datasets:
                if policy == "fail":
                    raise Exception(f"[{name}] Missing source: {s}")

        return {s: datasets[s] for s in sources if s in datasets}

    def process_spec(self, datasets, profile, spec):

        if ("name" not in spec) or ("description" not in spec):
            raise Exception("Invalid spec: name/description missing")

        name = spec["name"]
        handlers = self.get_handlers(profile)

        if (not isinstance(spec, dict)) or (len(spec) == 0):
            raise Exception("Spec should be a dict")

        if "generate" not in spec:
            # It is database operational specification..
            data = self.process_spec_default(datasets, profile, spec)
        else:
            # Custom callback
            data = self.process_spec_custom(datasets, profile, spec)

        # The spec processor can return multiple dataframes
        if isinstance(data, pd.DataFrame):
            data = {name: data}

        pipeline = spec.get("pipeline", None)
        if pipeline is not None:
            for processor in pipeline:
                if isinstance(processor, str):
                    if processor in handlers:
                        data = handlers[processor](data, spec)
                    elif hasattr(self, processor):
                        data = getattr(self, processor)(data, spec)
                    else:
                        raise Exception(f"Missing cleaner: {processor}")
                elif callable(processor):
                    data = processor(data, spec)
                else:
                    raise Exception(
                        "Only method names/callables are supported are supported"
                    )

        return

    def process_spec_custom(self, datasets, profile, spec):

        name = spec["name"]
        handlers = self.get_handlers(profile)

        # Custom...
        generate = spec["generate"]
        callback = None
        if (generate in handlers) and (not callable(handlers[generate])):
            callback = handlers[generate]
        elif hasattr(self, generate):
            callback = getattr(self, generate)

        if callback is None:
            raise Exception(f"[{name}] Invalid callback: {generate}")

        # Get hold of the data first...
        sources = self.get_spec_sources(spec, datasets)

        return callback(sources, spec)

    def process_spec_default(self, datasets, profile, spec):
        """
        Handle one specification at a time..
        """

        if ("dimensions" not in spec) or (not isinstance(spec["dimensions"], dict)):
            raise Exception("Dimensions in spec should be a dict")

        if ("metrics" not in spec) or (not isinstance(spec["metrics"], dict)):
            raise Exception("Metrics in spec should be a dict")

        # Get hold of the data first...
        sources = self.get_spec_sources(spec, datasets)

        if len(sources) > 1:
            raise Exception("Use custom spec handler for multiple sources")

        datasetdf = list(sources.values())[0]

        # now go through each of the dimensions
        dimensions = spec["dimensions"]
        metrics = spec["metrics"]

        _dfs = []
        for name, cols in dimensions.items():

            if isinstance(cols, str):
                cols = [cols]

            # Dont need to include other columns...
            relevant = cols + list(metrics.keys())
            df = datasetdf[relevant]

            # Check if there are lists and explode them...
            for col in cols:
                if isinstance(df.iloc[0][col], list):
                    df = df.explode(col)

            # Construct aggregates...
            df = df.groupby(cols)
            df = df.agg(metrics)

            # Clean up the index if multiple columns are specified
            if len(cols) > 1:
                df.index = df.index.map("+".join)
            df.index.name = "value"
            df = df.reset_index()

            # Also cleanup the column names...
            def clean_colname(what):
                if isinstance(what, (list, tuple)):
                    what = "_".join(what)
                    what = what.rstrip("_").lstrip("_")
                return what

            df.columns = df.columns.map(clean_colname)

            df.insert(0, "dimensions", name)

            _dfs.append(df)

        # merge all
        df = pd.concat(_dfs)
        del _dfs

        return {spec["name"]: df}

    def get_dataset_generic(self, source):
        """
        Use the dataset object to read the dataset
        """

        if (not hasattr(self, "read_data")) or (not hasattr(self, "get_dataset")):
            raise Exception(
                " get_dataset_raw expects read_data and get_dataset methods"
            )

        args = self.args
        start_date = self.args["start_date"]
        end_date = self.args["end_date"]

        name = source["name"]
        dataset = source["dataset"]
        params = source.get("params", {})
        setdate = source.get("date", False)

        cache = args.get("cache", False)
        cachename = f"{dataset}-{start_date}-{end_date}"
        cachefile = f"cache/{self.name}-cache-" + cachename + ".csv"

        if cache:
            try:
                os.makedirs(os.path.dirname(cachefile))
            except:
                pass
            if os.path.exists(cachefile):
                logger.debug(
                    "Read cached {}".format(name), extra={"transform": self.name}
                )

                df = pd.read_csv(cachefile, **params)
                return {name: df}

        datasetobj = self.get_dataset(dataset)

        df, metadata = datasetobj.read_data(
            start_date,
            end_date,
            filename=source["filename"],
            readfunc=self.read_data,
            params=params,
        )

        logger.debug("Read {}".format(name), extra={"transform": self.name})

        # Cache it for future use...
        if cache:
            df.to_csv(cachefile, index=False)

        # Insert lineage if possible
        lineage = None
        if ("files" in metadata) and (len(metadata["files"]) > 0):
            lineage = {
                "type": "lineage",
                "transform": self.name,
                "dependencies": [
                    {
                        "type": "file",
                        "nature": "input",
                        "objects": [metadata["files"][-1]],
                    },
                ],
            }

        self.update_frame(source, df, lineage)

        return {name: df}

    def process(self, state):
        """
        Run the computation and update the state
        """
        logger.debug(
            "Start execution", extra=self.config.get_extra({"transform": self.name})
        )

        # Will be used in other places..
        self.state = state

        profile = self.get_profile()

        # Get specs..
        specs = self.get_specs(profile)

        # First get the datasets
        datasets = self.get_datasets(profile, specs)

        # Now go through each spec and get the output..
        for spec in specs:
            enable = spec.get("enable", True)
            if not enable:
                continue
            self.process_spec(datasets, profile, spec)

        # Done
        logger.debug(
            "Complete execution", extra=self.config.get_extra({"transform": self.name})
        )

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        pass



import os
import sys
import copy
import json
import time
import numpy as np
from enrichsdk import Sink
import pandas as pd
from pandas.io import sql
import logging

from enrichsdk.lib.misc import *
from enrichsdk.core.frames import PandasDataFrame

logger = logging.getLogger("app")


class PQExport(Sink):
    """
    Parquet export for dataframes.

    The configuration requires a list of exports, each of which
    specifies a pattern for the frame name::

         'conf': {
            'args': {
                "exports": [
                  {
                      "name": "%(frame)s_pq",
                      "type": "pq", # optional. Default is pq
                      "frames": ["cars"],
                      "filename": "%(output)s/%(runid)s/%(frame)s.pq",
                      "params": {
                          # parquet parameters.
                          # "compression": 'gzip'
                          # "engine": 'auto'
                          # "index" :None,
                          # "partition_cols": None
                      }
                   }
                ]
            }
        }

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "PQExport"
        self.roles_supported = ["Export"]
        self.roles_current = "Export"

        data_root = os.path.join(os.environ["ENRICH_TEST"], self.name)
        self.testdata = {
            "data_root": data_root,
            "outputdir": os.path.join(data_root, "runs"),
            "inputdir": os.path.join(os.environ["ENRICH_TEST"]),
            "statedir": os.path.join(os.environ["ENRICH_TEST"], self.name, "state"),
            "conf": {
                "args": {
                    "exports": [
                        {
                            "name": "%(frame)s_pq",
                            # "type": "pq",
                            "frames": ["cars"],
                            "filename": "%(output)s/%(runid)s/%(frame)s.pq",
                            "sample": True,
                            "params": {
                                # "compression": 'gzip'
                                # "engine": 'auto'
                                # "index" :None,
                                # "partition_cols": None
                            },
                        }
                    ]
                }
            },
            "data": {
                "cars": {
                    "transform": "CarSales",
                    "filename": "sales.csv",
                    "params": {"sep": ","},
                }
            },
        }

    def preload_clean_args(self, args):
        """ """
        args = super().preload_clean_args(args)

        if len(args) == 0:
            raise Exception("Empty args provided")

        if ("exports" not in args) or (not isinstance(args["exports"], list)):
            raise Exception("PQExport requires a series of exports (a list)")

        for e in args["exports"]:

            if (
                (not isinstance(e, dict))
                or ("filename" not in e)
                or ("name" not in e)
                or ("frames" not in e)
            ):
                raise Exception(
                    "Each element of the export should be a dictionary with filename, type, and frames"
                )

            if e.get("type", "pq") not in ["pq"]:
                raise Exception("Only parquet exports are supported in current version")

        return args

    def process(self, state):
        """
        Export frames as parquet files as shown in the example.
        """

        # Sanity check...
        for e in self.args["exports"]:

            namebase = e["name"]
            params = e.get("params", {})
            sample = e.get("sample", True)

            # Collect all the frames data
            for f in e["frames"]:

                detail = state.get_frame(f)
                if detail is None:
                    raise Exception("Frame not present in state: {}".format(f))

                if detail["frametype"] != "pandas":
                    raise Exception("Frame not a pandas dataframe: {}".format(f))
                    continue

                df = detail["df"]

                # Resolve the locations
                filename = os.path.abspath(
                    self.config.get_file(e["filename"], extra={"frame": f})
                )
                relpath = self.config.get_relative_path(
                    filename, what="enrich_data_dir"
                )

                # Check over-rides
                overrides = self.frame_get_overrides(detail)
                save = overrides.get("save", True)
                if save:
                    try:
                        os.makedirs(os.path.dirname(filename))
                    except:
                        pass
                    df.to_parquet(filename, **params)

                    if sample:
                        size = min(1000, df.shape[0])
                        df.sample(size).to_parquet(filename + ".sample", **params)

                else:
                    logger.warn(
                        "Did not save {} due to overrides".format(f),
                        extra=self.config.get_extra(
                            {
                                "transform": self.name,
                                "data": "Overrides: {}".format(overrides),
                            }
                        ),
                    )

                if not os.path.exists(filename):
                    logger.error(
                        "PQ file not created or missing",
                        extra=self.config.get_extra(
                            {
                                "transform": self.name,
                                "data": "Filename: {}\nOverride Present: {}".format(
                                    filename, override_present
                                ),
                            }
                        ),
                    )
                    raise Exception("PQ file missing")

                # => Create state detail
                state_detail = {
                    "df": None,
                    "frametype": "db",
                    "transform": self.name,
                    "params": [
                        {
                            "filename": filename,
                            "action": "output",
                            "frametype": "binary",
                            "columns": self.collapse_columns(detail),
                            "descriptions": ["Parquet export of {} frame".format(f)],
                            "components": [
                                {
                                    "filename": relpath,
                                    "type": "pq",
                                    "rows": df.shape[0],
                                    "columns": df.shape[1],
                                    "sha256sum": get_checksum(filename),
                                    "filesize": "{0:0.1f} MB".format(
                                        get_file_size(filename) / (1024 * 1024)
                                    ),
                                    "modified_time": str(
                                        time.ctime(os.path.getmtime(filename))
                                    ),
                                    "create_time": str(
                                        time.ctime(os.path.getctime(filename))
                                    ),
                                }
                            ],
                        },
                        {
                            "type": "lineage",
                            "transform": self.name,
                            "dependencies": [
                                {
                                    "type": "dataframe",
                                    "nature": "input",
                                    "objects": [f],
                                },
                                {
                                    "type": "file",
                                    "nature": "output",
                                    "objects": [filename],
                                },
                            ],
                        },
                    ],
                    "history": [{"transform": self.name, "log": "Write PQite export"}],
                }
                try:
                    name = namebase % {"frame": f}
                    state.update_frame(name, state_detail, create=True)
                    state.make_note("Generated PQ export for {}".format(f))
                except:
                    logger.exception(
                        "Unable to store state",
                        extra=self.config.get_extra({"transform": self.name}),
                    )
                    raise Exception("Error while storing")

    def validate_results(self, what, state):
        """ """
        exports = self.args["exports"]
        for e in exports:
            namebase = e["name"]
            frames = e["frames"]
            for f in frames:
                name = namebase % {"frame": f}
                detail = state.get_frame(name)
                assert detail is not None


provider = PQExport



#!/usr/bin/env python3

"""
This script is a template for a prefect script. This can be
scheduled in a supervisor.

[program:REPO_THISSCRIPT]
command = /home/ubuntu/enrich/manage/bin/run_prefect_cmd.sh /home/ubuntu/enrich/customers/REPO/USECASE/workflows/prefect/THISSCRIPT.py ; Command to start app
user = ubuntu                                                         ; User to run as
group = ubuntu 
stdout_logfile = /home/ubuntu/enrich/logs/prefect/REPO_THISSCRIPT.log   ; Where to write log messages
redirect_stderr = true                                                ; Save stderr in the same log
environment=LANG=en_US.UTF-8,LC_ALL=en_US.UTF-8                       ; Set UTF-8 as default encoding
autostart=false
stopasgroup=true



"""
import os
import sys
import subprocess
import platform
import time
from datetime import date, timedelta
from prefect.engine.executors import LocalDaskExecutor

import prefect
from prefect import Flow, task
from prefect.schedules import CronSchedule
from prefect.tasks.github import GetRepoInfo
from prefect.triggers import any_failed
from prefect.utilities.logging import get_logger


def get_dates():
    daybefore = (date.today() + timedelta(days=-2)).isoformat()
    yesterday = (date.today() + timedelta(days=-1)).isoformat()
    tomorrow = (date.today() + timedelta(days=1)).isoformat()
    today = date.today().isoformat()
    return {
        "today": today,
        "yesterday": yesterday,
        "daybefore": daybefore,
        "tomorrow": tomorrow,
    }


@task
def run_pipeline(delay=0):
    logger = prefect.context.get("logger")
    logger.debug("Starting: Run Pipeline")

    dates = get_dates()
    if delay > 0:
        time.sleep(delay)

    cmd = [
        "/bin/bash",
        os.path.expandvars("$ENRICH_MANAGE/bin/run_enrichcli.sh"),
        "run-pipeline",
        "--scheduled",
        "--conf",
        os.path.expandvars(
            "$ENRICH_CUSTOMERS/acme/Marketing/pipelines/conf/PIPELINENAME.py"
        ),
        # Add required parameters
        # "TransformName:Variable=Value",
        # "SortEvents:backend=hs",
        "user=" + os.getlogin(),
    ]

    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = process.communicate()
    logger.debug("Output: {}".format(out))
    logger.debug("Error: {}".format(err))
    logger.debug("Completed: Run pipeline")


if __name__ == "__main__":
    with Flow("Running Enrich Pipeline", schedule=None) as flow:
        data = run_pipeline(delay=0)
        # A way to force the dependencies
        # data = run_another_pipeline(data)

    executor = LocalDaskExecutor(scheduler="processes")
    flow.run(executor=executor)



import os
import sys
import json
import hashlib
import pandas as pd
import random
from datetime import datetime, timedelta

from enrichsdk.datasets.discover import Dataset, DatabaseTable, \
           DatasetRegistry

def obfuscate(s):
    """
    Replace a string with  a hash
    """
    if ((s is None) or (pd.isnull(s))):
        return None
    if not isinstance(s, str):
        s = str(s)
    return hashlib.sha256(s.encode('utf-8')).hexdigest()[:16]

class AnonymizedDataset(Dataset):
    """
    Anonymize data before returning a sample
    """
    def sample(self, *args, **kwargs):
        safe = kwargs.get('safe', True)
        df = super().sample(*args, **kwargs)
        if not safe:
            return df

        # Anonymize additional columns...
        checkcols = ['alpha', 'beta']
        for c in df.columns:
            if len([check for check in checkcols if check in c]) > 0:
                df[c] = df[c]*random.randint(1,5)

        return df

datasets = []

############################################
# SQL-based datasets. These are typically generated by query execution
# engine, periodically based on pattern
############################################
for name,desc,sql in [
        ("DatasetName", "DatasetDescription", "SQLName")
]:
    d = AnonymizedDataset({
        "name": name,
        "alt_names": [sql],
        'description': desc,
        "paths": [
            {
                "nature": "s3",
                "name": "default",
                "path": "%(s3root)s/backup/%(node)s/rawdata/queries/" + sql
            },
        ],
        "match": {
            "generate": "generate_datetime_daily",
            "compare": "compare_datetime_pattern_match_range",
            "pattern": "%Y-%m-%d"
        },
        "subsets": [
            {
                "name": "Data",
                "pattern": ".*data.csv$"
            }
        ]
    })
    datasets.append(d)

############################################
# SQL-based datasets. These are typically generated by query execution
# engine, continuously replaced by the latest version..
############################################
for name,desc,sql in [
        ("ProviderCode","Provider Code", "provider_code"),
]:
    d = Dataset({
        "name": name,
        "alt_names": [sql],
        "description": desc,
        "paths": [
            {
                "nature": "s3",
                "name": "default",
                "path": "%(s3root)s/backup/%(node)s/rawdata/queries/" + sql
            },
        ],
        "match": {
            "generate": lambda start, end: [{ 'timestamp': 'complete', 'name': 'complete' }],
            "pattern": "*"
        },
        "subsets": [
            {
                "name": "Data",
                "pattern": ".*data.csv$"
            }
        ]
    })
    datasets.append(d)


############################################
# Feature Datasets. These are typically generated by pipelines. They
# will often have a version and timestamp as well..
############################################    
datasets.append(Dataset({
    "name": "DatasetName",
    "alt_names": ["cryptic_name"],
    "description": "Detailed description of the dataset",
    "paths": [
        {
            "nature": "s3",
            "name": "default",
            "path": "%(s3root)s/backup/%(node)s/data/customer/project/shared/datasetname/version/parameter"
        },
    ],
    "match": {
        "generate": "generate_datetime_daily",
        "compare": "compare_datetime_pattern_match_range",
        "pattern": "%Y-%m-%d"
    },
    "subsets": [
        {
            "name": "Sub-Dataset",
            "pattern": ".*sub-dataset.csv$"
        },
    ]
}))

def get_datasets(names=None):

    if names is not None and isinstance(names, list):
        return [ d for d in datasets if d.matches(names)]
    else:
        return datasets

def get_dataset(name):
    """
    Get the dataset object by name
    """
    for d in datasets:
        if d.matches(name):
            return d

    raise Exception("Unknown dataset: {}".format(name))

def get_registry(transform=None, state=None):

    registry = DatasetRegistry(transform=transform, state=state)
    registry.set_params({
        'enrich_data_dir': '/home/scribble/enrich/data',
        's3root': 'S3ROOT',
        'node': 'SERVERNAME',
        'remote_data_root': '/home/scribble/enrich'
    })

    registry.add_datasets(datasets)
    return registry



"""
Core classes that must be used to build transforms
"""
import os
import sys
import json
import copy
import re
import logging
import warnings
import markdown
import importlib
import traceback
import s3fs
import pandas as pd
from datetime import datetime
import abc
import jsonschema

from ..lib.exceptions import *
from ..lib import renderlib
from ..lib.misc import get_checksum, get_file_size

logger = logging.getLogger("app")
thisdir = os.path.dirname(__file__)
templatedir = os.path.join(thisdir, "..", "templates")
templatedir = os.path.abspath(templatedir)
widgetdir = os.path.join(templatedir, "widgets")


__all__ = ["Node", "Transform", "Model", "Skin", "Integration"]


def flatten(c):
    if isinstance(c, (int, float, str)):
        return c

    try:
        if isinstance(c, (datetime, date)):
            c = c.isoformat()
        elif isinstance(c, (dict, list)):
            c = json.dumps(c)
        else:
            c = str(c)
    except:
        c = str(c)
    return c[:30]


class NodeMeta(abc.ABCMeta):
    """
    Meta class for all elements with schemas. This allows for
    registration, validation, and tracking of the schema implementors.
    """

    def __init__(cls, name, bases, dct):
        if cls.__name__ not in ["Node"]:
            if not hasattr(Node, "_registry"):
                Node._registry = []

            if cls.instantiable():
                Node._registry.append(cls)

        # Now initialize
        super().__init__(name, bases, dct)


##################################################
# Transform objects
##################################################
class Node(metaclass=NodeMeta):
    """
    This is the base class for all modules including transforms,
    models, skins etc.

    This provides a number of infrastructural elements that enable the
    module to the plugged into the application infrastructure.

    This is for understanding only. Subclassing is done from derived
    classes such as Source, SearchSkin etc.
    """

    @classmethod
    def implementation_list(cls):
        return cls._registry

    @classmethod
    def implementation_get(cls, name):
        for c in cls._registry:
            if c.__name__ == name:
                return c

        raise InvalidNode("Unknown node name: {}".format(name))

    @classmethod
    def instantiable(cls):
        """
        Whether this transform should/can be instantiated.
        This is usually used by base classes.
        """
        return True

    def __init__(self, *args, **kwargs):

        # Note the this or subclass path path
        m = importlib.import_module(self.__module__)
        self.transform_root = os.path.abspath(os.path.dirname(m.__file__))

        if "config" not in kwargs:
            raise Exception("config parameter is missing")

        self.config = kwargs.pop("config")
        """
        object: Configuration object

        This is passed to the module during instantiation by the
        enrich compute engine
        """
        self.node_type = "Unknown"
        """
        str: Node type

        This is one of Transform, Model, Skin, Integration
        """

        self.name = "Unknown"
        """
        str: Name of the transform. Required

        This is overridden by name specified in the manifest.json
        """

        self.description = "Unknown"
        """
        str: Text description of the transform

        This is overridden by description specified in the manifest.json
        """

        self.author = "Builtin"
        """
        str: Author of this module.

        This is overridden by author specified in the manifest.json
        """

        self.version = 1.0
        """
        float: Version of this module
        """

        self.minorversion = 1.0
        """
        float: Minor version of this module
        """

        self.output_version = 1.0
        """
        float: Version of the output of this module
        """

        self.debug = False
        """
        bool: Debug mode operation at the transform level
        """

        self.files = ["manifest.json"]
        """
        list: Files in this module
        """

        self.test = False
        """
        bool: Whether this transform is being operated in test mode
        """

        self.dependencies = {}
        """Specifies the list of frames and transformations that this
        transform dependens on. This value has to be overriden by the
        subclass explicitly. Otherwise it transform will throw a
        validation error. For example::

            {
              'article': 'Core',
              'store': ['EnrichedStore', 'EnrichedTransform']
              ...
            }

        """

        self.datasets = []
        """
        Input and output datasets handled by this transform
        """

        self.args_schema = {}
        """
        dict: JSON Schema to validate the input args::

           See: https://python-jsonschema.readthedocs.io/en/latest/
                https://json-schema.org/

            self.args_schema = {
                "type": "object",
                "properties": {
                    "name": {
                         "type": "string"
                         "description": "Name of the dataset"
                     },
                    "count": {
                         "type": "number"
                         "description": "Threshold"
                     }
                     ...
                 }
                 "required": ["name", "count"]
                }
            }

        """

        self.supported_extra_args = []
        """
        Extra arguments that can be passed on the command line and the GUI to
        the module::

            self.supported_extra_args = [
                {
                    "name": "jobid",
                    "description": "Export JobID",
                    "type": "str",
                    "required": True,
                    "default": "22832"

                }

            ]

        """

        self.data_root = self.get_file("%(data_root)s")
        """

        """
        self.testdata = self.get_testdata
        """dict: Test data to be used for testing module

        The default structure is instantiated by self.get_testdata. Each
        of the elements (e.g., conf) can be specified by over-riding
        corresponding method (e.g., self.get_testdata_conf). Such
        functions are supported for conf, args, global, data.

        It has the form::

            {
              'data_root': os.path.join(os.environ['ENRICH_TEST'], self.name),
              'outputdir': os.path.join(os.environ['ENRICH_TEST'], self.name),
              'inputdir': os.path.join(os.environ['ENRICH_TEST']),
              'statedir': os.path.join(os.environ['ENRICH_TEST'], self.name, 'state'),
              'datasets': {
               ...
              },
              'global': {
                'args': {
                    'rundate': '2020-01-01'
                }
              },
              'conf': {
                 'args': {
                       'threshold': 0.6
                       ...
                  }
              },
              'data': {
                'article': {
                    'filename': 'test.csv',
                    'transform': 'Core',
                    'frametype': 'pandas',
                    'stages': [ 'transform1'],
                    'params': {
                        'sep': ','
                    }
                }
              }
            }

        When the module is tested, the sdk loads
        `$ENRICH_DATA/test/temp-configname/Core/test.csv` and makes it
        available to the module using the state interface

        If data in the frame is to be loaded from multiple files,
        you can specify the list and a merge function that gets
        the list of frames to merge::

            {
              ..
              'data': {
                'article': {
                    'filename': {
                         'transform1': ['test.csv'],
                         'transform2': ['test.csv']
                    },
                    'frametype': 'pandas',
                    'mergedf': self.test_merge_func, # function
                    'transform': 'transform1',
                    'stages': ['transform2'],
                    'params': {
                        'sep': ','
                    }
                }
              }
            }
            ...

           def test_merge_func(self, dfs):
               ...

        See `get_test_conf` function to dynamically generate these
        configurations.

        `datasets` is a dictionary with a specification for where the
        input datasets should be stored and obtained from::

            {
               "command": command,
               'params': {
                   'enrich_data_dir': '/home/ubuntu/enrich/data',
                   'backup_root': 'some-s3-path',
                   'node': 'some hostname'
                },
               'available': [
                    Dataset({
                       'name': "inventory_dataset",
                       ...
                    }),
                     ...
               ]
            }

        """

        self.enable = True
        """
        bool: Is this transform enabled?

        A transform could be included but not enabled
        """

        self.outputs = {}
        """
        dict: Outputs of this module

        It has the form::

            {
                <pattern> : <description>,
                <pattern>: {
                   'description': <text>
                }
            }
            """
        self.tags = []
        """
        list: Tags associated with this transform
        """

        self.metadata = {}

        self.data_version_map = {}
        """
        dict: Versions of the preprocessed data

        .. warning::

            .. deprecated:: 2.0

        It has the form::

            {
               'raw': 1.0,
               'preprocessed': 1.3
            }
        """

        self.roles_supported = ["default"]
        """
        list: A transform can support multiple roles such as source, sink etc. This lists the roles.

        This attribute was used to sequence the transforms in early versions of Enrich. Now it is deprecated.

        .. warning::

            .. deprecated:: 2.0


        """

        self.roles_current = "default"
        """
        str: Current role of this transform

        This attribute was used to sequence the transforms in early versions of Enrich. Now it is deprecated.

        .. warning::

           .. deprecated:: 2.0

        """

        # Interface variables...
        self.manifest = {}
        """
        dict: Manifest json file loaded
        """

        self.args = {}
        """
        dict: Arguments passed to the transformation from the configuration file

        """

        try:
            module = sys.modules[self.__module__]
        except:
            raise Exception("Could not find module")

        self.modulepath = os.path.dirname(module.__file__)

    @property
    def fullname(self):
        """
        str: One line summary of the transform
        """
        return "{} (v{}:v{} by {})".format(
            self.name, self.version, self.minorversion, self.author
        )

    #############################################
    # Helper functions
    #############################################
    def get_testdata(self):
        """
        Generate test data structure

        Returns:

           testdata (dict) : Dictionary

        """
        test_root = os.environ["ENRICH_TEST"]

        return {
            "data_root": os.path.join(test_root, self.name),
            "inputdir": test_root,
            "outputdir": os.path.join(test_root, self.name, "output"),
            "statedir": os.path.join(test_root, self.name, "state"),
            "conf": self.get_testdata_conf(),
            "data": self.get_testdata_data(),
            "global": self.get_testdata_global(),
        }

    def get_testdata_conf(self):
        """
        Get testdata conf attribute

        Returns
        -------
        dict: a dictionary of args and any other elements

        """
        return {"args": self.get_testdata_args()}

    def get_testdata_args(self):
        """
        Get testdata args attribute of conf

        Returns
        -------
        dict: a dictionary of args

        """
        return {}

    def get_testdata_global(self):
        """
        Get testdata global attribute. Global args are pipeline-wide
        args.

        Returns
        -------
        dict: a dictionary of global args

        Example::

            return {
               'args': {
                    'rundate': '2020-01-01'
                }
             }

        """
        return {}

    def get_testdata_data(self):
        """
        Get 'data' element of test data. This specifies
        what dataframes have to be loaded to test this
        transform.

        Returns
        -------
        dict: A specification of dataframes and loading instructions

        Example::

          return {
                'article': {
                    'filename': 'test.csv',
                    'transform': 'Core',
                    'frametype': 'pandas',
                    'stages': [ 'transform1'],
                    'params': {
                        'sep': ','
                    }
                }
              }
        """
        return {}

    def is_enabled(self):
        """
        bool: Check whether this module is enabled

        Returns
        -------
        bool
        """
        return self.enable

    def is_transform(self):
        """
        bool: Test whether the module is a transform

        Returns
        -------
        bool
        """
        return self.node_type == "Transform"

    def is_model(self):
        """
        Test whether the module is a module

        Returns
        -------
        bool
        """
        return self.node_type == "Model"

    def is_skin(self):
        """
        Test whether the module is a skin

        Returns
        -------
        bool

        .. warning::

           .. deprecated:: 2.0
           Use enrichsdk.render library functions

        """
        return self.node_type == "Skin"

    def is_integration(self):
        """
        Test whether the module is an integration

        Returns
        -------
        bool

        .. warning::

           .. deprecated:: 2.0
           Will be dropped in future

        """
        return self.node_type == "Integration"

    def is_sink(self):
        """
        Is current role a sink?

        Returns
        -------
        bool

        """
        return (self.node_type == "Transform") and (self.roles_current == "Sink")

    def is_source(self):
        """
        Is current role a source?

        Returns
        -------
        bool

        """
        return (self.node_type in ["Transform", "Model"]) and (
            self.roles_current == "Source"
        )

    def has_tag(self, tags):
        """
        Does the module have any of the specified tags?

        Returns
        ----
        bool
            True if there is an overlapping tag otherwise False

        """
        common = [t for t in tags if t in self.tags]
        return len(common) > 0

    def is_search(self):
        """
        Is the current module a search skin?

        Returns
        -------
        bool
        """
        return (self.node_type == "Skin") and self.has_tag(["search"])

    def is_notification(self):
        """
        Is the current module a notification implementation?

        Returns
        -------
        bool
        """
        return (self.node_type == "Integration") and self.has_tag(["notification"])

    def get_supported_extra_args(self):
        """
        Get the command line arguments. This will handle any callables specified
        by the default.

        Returns:
           args (list): List of arg specifications. Default values can be actual   strings/numbers or callback functions

        See supported_extra_args
        """

        final = []
        for a in self.supported_extra_args:
            a = copy.copy(a)
            if callable(a["default"]):
                a["default"] = a["default"]()
            final.append(a)

        return final

    #################################################
    # Configuration the Module
    #################################################
    def preload_clean_args(self, args):
        """
        Clean the arguments before using them

        Args:
           args: args to be resolved/cleaned/extended

        Returns:
           list: Cleaned args

        """

        # Nothing to clean if there are no extra args that could be
        # changed at runtime.
        if not self.config.enable_extra_args:
            return args

        # Override the defaults with the parameters
        # Three sources for a given args attribute in the order of
        # priority
        #
        # 1. command line
        # 2. Configuration
        # 3. Default

        transform = self.name
        readonly = self.config.readonly
        cmdargs = self.config.get_cmdline_args()

        # Module-specific args..
        supported_extra_args = self.get_supported_extra_args()

        # Insert the global args if not already present. Module
        # specific value overrides the global values
        duplicates = []
        present = [a["name"] for a in supported_extra_args]
        for arg in self.config.supported_extra_args:
            arg = copy.copy(arg)
            if arg["name"] not in present:
                supported_extra_args.append(arg)
                present.append(arg["name"])
            else:
                duplicates.append(arg["name"])

        if len(duplicates) > 0:
            msg = "Module-specific args over-riding the global args\n"
            msg += "Args : {}".format(", ".join(duplicates))
            logger.warning(
                "Module args over-riding the global args",
                extra={"transform": self.name, "data": msg},
            )

        # Collect the final values for all the defaults
        final = {}
        for a in supported_extra_args:
            name = a["name"]
            transform_cmdlabel = "{}:{}".format(transform, name)
            global_cmdlabel = "global:{}".format(name)
            default = a["default"]
            if callable(default):
                default = default()
            required = a["required"]
            if required and not readonly:
                # Required and specified on the command line...
                if transform_cmdlabel in cmdargs:
                    final[name] = cmdargs[transform_cmdlabel]
                elif global_cmdlabel in cmdargs:
                    final[name] = cmdargs[global_cmdlabel]
                elif name in args:
                    final[name] = args[name]
                else:
                    logger.error(
                        "Invalid configuration",
                        extra={
                            "transform": self.name,
                            "data": "Variable {} for transform {} is required but missing in command line".format(
                                name, self.name
                            ),
                        },
                    )
                    raise Exception(
                        "Missing Commandline arg: {} or {}".format(
                            transform_cmdlabel, global_cmdlabel
                        )
                    )
            elif transform_cmdlabel in cmdargs:
                # Optional but specified on the command line
                final[name] = cmdargs[transform_cmdlabel]
            elif global_cmdlabel in cmdargs:
                # Optional but specified on the command line
                final[name] = cmdargs[global_cmdlabel]
            elif name in args:
                # Priority 2
                final[name] = args[name]
            else:
                # Priority 3
                final[name] = default

        # => Log what has happened...
        msg = ""
        for k, v in final.items():
            msg += "{}: {}\n".format(k, v)
        logger.debug(
            "Configuration has been overridden",
            extra={"transform": self.name, "data": msg},
        )

        args.update(final)

        return args

    def preload_validate_conf(self, conf):
        """
        Check whether this configuration is even valid?

        Args:
           conf: Module configuration specified in the config file

        Raises:
            Generic exception if conf is not a dictionary or if the version doesnt match

        """

        if not (isinstance(conf, dict) and len(conf) > 0):
            logger.error(
                "Conf is not a valid dictionary",
                extra=self.config.get_extra({"transform": self.name}),
            )
            raise Exception("Conf is not a valid dictionary")

        if self.version != conf.get("version", 1.0):
            logger.error(
                "Version mismatch",
                extra=self.config.get_extra({"transform": self.name}),
            )
            raise Exception("Version mismatch between config and module")

    def preload_validate_self(self):
        """
        Syntactic check of the supported_extra_args, output
        other future configurations
        """
        supported_extra_args = self.supported_extra_args
        if not isinstance(supported_extra_args, list):
            logger.error(
                "Invalid configuration. Expected {}, actual {}".format(
                    "list", str(type(supported_extra_args))
                ),
                EXTRA=self.config.get_extra(
                    {
                        "transform": self.name,
                    }
                ),
            )
            raise Exception("Supported extra argument should be a list of dicts")

        for i, a in enumerate(supported_extra_args):
            if not isinstance(a, dict):
                logger.error(
                    "Invalid configuration. Entry {} of supported_extra_args: Expected {}, actual {}".format(
                        i, "dict", str(type(a))
                    ),
                    EXTRA=self.config.get_extra(
                        {
                            "transform": self.name,
                        }
                    ),
                )
                raise Exception("Supported extra argument should be a list of dicts")

            required = [
                "name",
                "description",
                "default",
                "required",
            ]
            missing = [x for x in required if x not in a]
            if len(missing) > 0:
                logger.error(
                    "Invalid configuration. Entry {} of supported_extra_args: Missing attributes: {}".format(
                        i, ", ".join(missing)
                    ),
                    extra=self.config.get_extra(
                        {"transform": self.name, "data": str(a)}
                    ),
                )
                raise Exception("Missing attributes in one of the supported extra args")

        if not hasattr(self, "outputs"):
            raise Exception("Output definition is missing in transform. Critical error")

        if not isinstance(self.outputs, dict):
            raise Exception("Output definition (self.outputs) should be a dict")

        for k, v in self.outputs.items():
            if not isinstance(k, str):
                raise Exception(
                    "Output definition (self.outputs) should have string frame names only ({})".format(
                        k
                    )
                )

            if not isinstance(v, dict):
                raise Exception(
                    "Output definition (self.outputs) should have dict for each frame name ({})".format(
                        k
                    )
                )

    def configure(self, conf):
        """
        Load the module and prepare to execute

        Args:
           conf: Module configuration specified in config

        """
        readonly = self.config.readonly

        # => Check whether internal configuration is valid
        self.preload_validate_self()

        # Validate the configuration specified
        self.preload_validate_conf(conf)

        self.conf = copy.deepcopy(conf)

        if "name" in conf:
            self.name = conf["name"]

        if "output_version" in conf:
            self.output_version = conf["output_version"]

        if "test" in conf:
            self.test = True

        if "enable" in conf:
            self.enable = conf["enable"]

        if "debug" in conf:
            self.debug = conf["debug"]

        if "role" in conf:
            self.roles_current = conf["role"]

        # Add a readonly check to avoid side effects of the enrichgui
        # scanning the transform.
        if "args" in conf:
            if not readonly:
                self.args = self.preload_clean_args(conf["args"])

        if "dependencies" in conf:
            self.dependencies = conf["dependencies"]

    #####################################################
    # => Testing
    #####################################################
    def get_test_conf(self):
        """
        Function to get test configurations.

        This can be used to dynamically generate
        test configurations by overiding this function
        in the derived class.
        """
        return self.testdata.get("conf", {})

    #####################################################
    # => Validation.
    #####################################################
    def validate(self, what, state):
        """
        Validate various aspects of the transform state,
        configuration, and data. Dont override this function. Override
        specific functions such as `validate_args`.

        Args:
            what (str): What should be validated? args, conf, results etc.
            state (object): State to checked

        Returns:

             Nothing

        Raises:

             Exception ("Validation error")

        """

        func = getattr(self, "validate_" + what, None)
        if hasattr(func, "__call__"):
            return func(what, state)
        else:
            logger.error(
                "Cannot find function to validate {}".format(what),
                extra=self.config.get_extra({"transform": self.name}),
            )
            raise Exception("Validation error")

    def validate_conf(self, what, state):
        """
        Validate user-specified configuration for both content and
        structure.

        Args:
            what (str): "conf"
            state (object): State to checked

        Returns:

             Nothing

        Raises:

             Exception ("Validation error")

        """
        if self.name in [None, "Unknown"]:
            raise Exception("Invalid name for transform")

        if self.node_type == "Unknown":
            raise Exception("Invalid node type")

        if not isinstance(self.dependencies, dict):
            raise Exception("Dependencies should be a dictionary")

    def validate_args(self, what, state):
        """
        Validate user-specified args for both content and
        structure against args_schema.

        Args:
            what (str): "args"
            state (object): State to checked

        Returns:

             Nothing

        Raises:

             Exception ("Validation error")

        """
        # Validate the schema
        try:
            jsonschema.validate(self.args, self.args_schema)
        except:
            logger.exception(
                "Invalid/insufficient arguments. Doesnt match schema",
                extra={"transform": self.name},
            )

    def validate_input(self, what, state):
        """
        Validate input data including dataframes in
        the state

        Args:
            what (str): "args"
            state (object): State to checked

        Returns:

             Nothing

        Raises:

             Exception ("Validation error")

        """
        pass

    def validate_results(self, what, state):
        """
        Post-processing validation check to make sure that the
        computation has happened correctly - existance and values
        of the results.

        Args:
            what (str): "results"
            state (object): State to checked

        Returns:

             Nothing

        Raises:

             Exception ("Validation error")

        """

        pass

    def validate_testdata(self, what, state):
        """
        Validate the test data provided for structure and
        semantics. Test for existance of files and appropriate load
        commands being present.

        Args:
            what (str): "testdata"
            state (object): State to checked (generally ignored)

        Returns:

             Nothing

        Raises:

             Exception ("Validation error")
        """
        if (
            (not hasattr(self, "testdata"))
            or (self.testdata is None)
            or (not isinstance(self.testdata, dict))
        ):
            raise Exception("Invalid test data. Either is it None or not a dictionary")

        data = self.testdata.get("data", {})
        for framename, detail in data.items():
            if "filename" not in detail:
                raise Exception(
                    "Missing frame detail: filename. Only file-based test input supported for now"
                )
            if "params" not in detail:
                raise Exception("Missing frame detail: params")
            if not isinstance(detail["params"], dict):
                raise Exception("Invalid frame detail: params should be a dictionary")
            frametype = detail.get("frametype", "pandas")
            if (frametype in ["pandas", "spark"]) and ("sep" not in detail["params"]):
                raise Exception(
                    "Missing frame detail: A separator 'sep' is minimum required parameter"
                )

        conf = self.testdata.get("conf", {})
        if not isinstance(conf, dict):
            raise Exception("Invalid test configuration: conf should be a dictionary")

    #####################################################
    # => Initialize the system...
    #####################################################
    def initialize(self):
        """
        Called at the time of instantiating the module.

        This can be used to open connections, setup state and other
        functions.

        .. deprecated::

           Use configure/other functions to complete the initialization..

        """
        pass

    #####################################################
    # => Load the module
    #####################################################
    def load(self):
        """
        Run time loading of parameters.

        The primary use of this function is for skins. They load the
        last run details for further processing.
        """
        pass

    #####################################################
    # => Check whether preconditions have been met
    #####################################################
    def ready(self, state):
        """
        Test whether the module is ready for processing.

        This is typically not over-ridden. The platform implemets a
        number of checks including preconditions and checks before
        calling the module. The developer specifies the preconditions
        using the dependencies object (which could be overridden in
        the configuration)

        This function is there for special cases.
        """
        return True

    #####################################################
    # => Execute
    #####################################################
    def process(self, state):
        """
        Execute the module function whatever it is.

        The process function typically extracts the dataframes from
        the state, computes on it, and writes them back.

        Every module implements this function except Skins.

        Args:
           state: State object passed by the pipeline manager

        """
        pass

    ########################################
    # Helper functions
    ########################################
    def get_credentials(self, name):
        """
        Looks up the siteconf in ENRICH_ETC and
        returns appropriate entry

        This name should in specific in the credentials
        section of siteconf.


        """
        if not hasattr(self.config, "siteconf"):
            raise Exception("No siteconf")

        siteconf = self.config.siteconf
        if not isinstance(siteconf, dict):
            raise Exception("Invalid siteconf format")

        try:
            credentials = siteconf["credentials"][name]
        except:
            logger.exception(
                "Cannot acquire credentials",
                extra=self.config.get_extra({"transform": self.name}),
            )
            raise Exception("Missing/invalid credentials")

        return credentials

    def get_versionmap(self, include_tags=False):
        """
        Export the versionmap

        Args:
            include_tags (bool): Whether to include git tags

        Returns:
            verionmap (list): List of dicts - one for each module

        """
        if not hasattr(self.config, "versionmap"):
            raise Exception("No versionmap")

        versionmap = copy.deepcopy(self.config.get_versionmap())

        # Drop tags unless specified explicitly
        if not include_tags:
            for v in versionmap:
                v.pop("alltags", [])

        return versionmap

    def get_default_metadata(self, state):
        """
        Get reuse metadata dict with some standard fields.

        Args:
            state (object): State object passed by the pipeline

        Returns:
            metadata (dict): Dictionary with standalone:pipeline schema

        """

        usecase = self.config.get_usecase()

        versionmap = self.get_versionmap()
        for v in versionmap:
            v.pop("branches", None)

        metadata = {
            "schema": "standalone:pipeline",
            "version": "1.0",
            "timestamp": datetime.now().replace(microsecond=0).isoformat(),
            "pipeline": {
                "customer": usecase,
                "usecase": usecase,
                "name": state.name,
                "description": self.config.description,
                "host": state.stats["platform"]["node"],
                "runid": state.runid,
                "pid": state.pid,
                "start_time": state.start_time.replace(microsecond=0).isoformat(),
                "end_time": state.end_time.replace(microsecond=0).isoformat(),
                "versionmap": versionmap,
                "stats": state.stats,
                "cmdline": list(sys.argv),
            },
        }

        return metadata

    #############################################
    # Helper functions
    #############################################
    def get_file(self, *args, **kwargs):
        """
        File path resolver. Insert transform and pass to the pipeline
        execution engine.

        Args:
           abspath (bool): Generate absolute path (default: True)
           extra (dict): Extra args for resolving the path

        Returns:
           path (str): Resolved path

        """

        extra = kwargs.pop("extra", {})
        extra = copy.copy(extra)
        if "transform" not in extra:
            extra["transform"] = self

        # Insert caller information?
        extra["transform_root"] = self.transform_root

        # Pass it back
        kwargs["extra"] = extra

        return self.config.get_file(*args, **kwargs)

    def get_relative_path(self, path, what="enrich_data_dir"):
        """
        Get the path relative to a predefined root

        Args:
             path (str): Path of object
             what (str): What should the root of the relative path be?

        """
        return self.config.get_relative_path(path, what)

    def note_access(self, filename, *args, **kwargs):
        """
        Note access to file to enable creation of lineage

        Args:
             filename (str): File being accessed
             args (list): Extra args to be passed to baseclass for use
        """
        self.config.note_access(filename, self, *args, **kwargs)

    def get_cache_dir(self, name=None, subdir=".", what="raw", create=False, extra={}):
        """
        Get a local directory for caching partial results.

        .. warning::

            Deprecated. Not to be used.

        Args:
             name (str): Namespace
             version (str): Version of the cache
             what (str): Class of data being stored
             create (bool): Create the cache directory
             extra (dict): Extra parameters passed to get_file

        Returns:
             path (str): Path of the cache directory

        """
        warnings.warn("Function will be removed soon", DeprecationWarning)

        if name is None:
            name = self.name

        try:
            version = self.data_version_map[what][name]
        except:
            version = 1.0

        if "cache" not in self.args:
            raise Exception("Cache directory not specified")

        cache_path = self.config.get_file(self.args["cache"], extra=extra)

        path = os.path.join(cache_path, what, name, str(version), subdir)
        path = os.path.abspath(path)
        if create:
            try:
                os.makedirs(path)
            except:
                pass

        return path

    def get_arg_overrides(self, state):
        """
        Get any over-rides provided from other transforms.

        Args:
           state (object): State object

        Returns:
            overrides (list)

        """
        return state.transform_overrides.get(self.name, [])

    def set_arg_overrides(self, state, name, detail):
        """
        Set any over-rides provided from other transforms.
        """

        if name not in state.transform_overrides:
            state.transform_overrides[name] = []
        state.transform_overrides[name].append(detail)

    def pass_args(self, state, name, detail):
        """
        Alias for set_arg_overrides
        """
        return self.set_arg_overrides(state, name, detail)

    def frame_get_overrides(self, state_detail):
        """
        Obtain any extra instructions passed by previous
        stage on how to process a frame

        Returns
        -------
            args (dict): Dictionary passed by previous transform

        """

        default = {}
        if (not isinstance(state_detail, dict)) or ("params" not in state_detail):
            return default

        params = state_detail["params"]
        if isinstance(params, dict):
            params = [params]
        for p in params:
            # Is this params of the type args?
            if p.get("type", "unknown") != "overrides":
                continue
            # Is this meant for me?
            if p.get("transform", "Unknown") != self.name:
                continue
            # Return args
            return p.get("args", {})

        return default


##################################################
# Skin base class
##################################################
class Skin(Node):
    """

    .. deprecated:: 2.0

    Renders the data computed the transforms and modules.

    This is typically used by built-in capabilities such as Search.
    An alternative to implementing this is to use the usecase apps
    capabilities (explained elsewhere)

    """

    def __init__(self, *args, **kwargs):
        super(Skin, self).__init__(*args, **kwargs)
        self.node_type = "Skin"

    # Template rendering functions
    def template_render(self, widgetname, context):
        """
        Render a specified template using the context

        Args:
            widgetname (str): Name of the template
            context (dict): Key-value pairs

        Returns:
            Rendered html template that can be embedded

        """
        return renderlib.template_render(widgetdir, widgetname, context)

    def template_get_variables(self, widgetname):
        """
        Get the variables from a template

        Args:
            widgetname (str): Name of the template

        Returns:
            List of variables
        """
        return renderlib.template_get_variables(widgetdir, widgetname)

    def load(self):
        """
        Run time loading of parameters.

        Each skin loads the last run details for further processing.
        """
        pass

    def render(self):
        """
        This is a generic data rendering capability. Each
        implementation of a skin should provide this.

        This generic rendering capability inturn calls implementation
        specific rendering called render_helper (see below)

        """
        raise Unsupported("Render should be implemented by the plugin")


##################################################
# Transform objects
##################################################
class Transform(Node):
    """
    This is the base class for all transforms.
    """

    def __init__(self, *args, **kwargs):
        super(Transform, self).__init__(*args, **kwargs)
        self.node_type = "Transform"

    def configure(self, conf):

        super(Transform, self).configure(conf)

    def collapse_columns(self, details):
        """
        Collect all columns
        """
        return {}

    #####################################################
    # => Generate state management information
    #####################################################
    def get_file_params(self, name, df, path, fshandle=None):
        """
        Generate output metadata for files (required for compliance)

        Args:
            df (Dataframe): Pandas dataframe
            path (str): Path where the dataframe was dumped
            fshandle (obj): S3 filesystem handle of (s3fs)

        Returns:
            list: List of dicts

        """

        # Collect file information for files that are written to s3/filesystem
        if fshandle is not None:
            info = fshandle.info(path)
            component = {
                "filename": path,
                "sha256sum": info["ETag"].replace('"', ""),
                "filesize": "{0:0.3f} MB".format(info["size"] / (1024 * 1024)),
                "modified_time": info["LastModified"].ctime(),
                "create_time": info["LastModified"].ctime(),
            }
        else:
            root = self.config.enrich_data_dir
            component = {
                "filename": os.path.relpath(path, root),
                "sha256sum": get_checksum(path),
                "filesize": "{0:0.3f} MB".format(get_pathsize(path) / (1024 * 1024)),
                "modified_time": str(time.ctime(os.path.getmtime(path))),
                "create_time": str(time.ctime(os.path.getctime(path))),
            }

        # Insert shape as well..
        component.update({"rows": df.shape[0], "columns": df.shape[1]})

        return [
            {
                "action": "output",
                "type": "table",
                "frametype": "pandas",
                "filename": path,
                "components": [component],
                "columns": self.get_column_metadata(name, df),
                "ddl": pd.io.sql.get_schema(df, name),
            }
        ]

    def get_column_params(self, name, df):
        """
        Generate columns metadata to be associated with
        this dataframe in the state. In params form.

        Args:
            name (str): name of the dataframe being processed
            df (DataFrame): Dataframe to be documented

        Returns:
            List of dicts. Each dict has 'type' (columns) and
            columns metadata (a dict)
        """
        columns = self.get_column_metadata(name, df)

        return [{"type": "compute", "columns": columns}]

    def get_column_metadata(self, name, df):
        """
        Generate columns metadata to be associated with
        this dataframe in the state.

        Args:
            name (str): name of the dataframe being processed
            df (DataFrame): Dataframe to be documented

        Returns:
            Dict with columns metadata
        """

        sampledf = df.sample(min(10, df.shape[0]))

        columns = {}
        for index, c in enumerate(df.columns):

            # Handle the case where there are duplicate column names
            sample = [flatten(c) for c in sampledf.iloc[:, index].tolist()]

            # Handle the case where c is a complex column name
            c_slug = c
            if not isinstance(c_slug, (str, int)):
                c_slug = str(c_slug)

            columns[c_slug] = {
                "touch": self.name,  # Who is introducing this column
                "datatype": df.get_generic_dtype(c),  # What is its type
                "description": self.get_column_description(name, c),
                "version": self.get_column_version(name, c),
                "security": self.get_column_security(name, c),
                "sample": sample,
            }

        return columns

    #####################################################
    # => Column output
    #####################################################
    def lookup_column_output(self, frame, column, outputs=None):
        """
        Look through the output definition to extract the
        description. The output description sometimes specifies a
        pattern that must be matched. Invalid patterns or patterns
        that dont match are ignored.

        Args:
            frame (str): name of the dataframe being processed
            column (str): Name of the column whose description is
                        required

            outputs (dict): A dictionary that has a set of patterns for each frame.

        Returns:
            Description if found else empty string
        """

        # If not over-ride is present, then use the
        # outputs passed in the parameter
        if outputs is None:
            outputs = self.outputs
            if frame not in outputs:
                return []
            outputs = outputs[frame]

        # => Do the exact match first..
        if column in outputs:
            detail = copy.copy(outputs[column])
            if isinstance(detail, str):
                return [{"description": detail}]
            elif isinstance(detail, dict):
                if "description" not in detail:
                    detail["description"] = ""
                return [detail]
            else:
                raise Exception("Unexpected output definition")

        # Now do a regular expression match
        # First sort all patterns by length
        patterns = sorted(list(outputs.keys()), key=lambda k: len(k), reverse=True)

        selected = []
        for pattern in patterns:

            # Get the detail..
            detail = copy.copy(outputs[pattern])
            if isinstance(detail, str):
                """
                If a string is specified, then that is the description
                """
                detail = {"description": detail}

            # Get the regular expression
            matcher_re = detail.get("matcher", r"^{}$".format(pattern))

            try:
                matcher = re.search(matcher_re, column, flags=re.IGNORECASE)
                if matcher is None:
                    continue
            except:
                continue

            # Capture the regex matcher as well. It may have variables
            # from the column name. This is an advanced usecase.
            detail["matcher"] = matcher

            # Return a dictionary...
            selected.append(detail)

        return selected

    def get_column_description(self, frame, column, outputs=None):
        """
        Look through the output definition to extract the
        description. The output description sometimes specifies a
        pattern that must be matched. Invalid patterns or patterns
        that dont match are ignored.

        Args:
            frame (object): name of the dataframe being processed
            column (object): Name of the column whose description is
                        required

        Returns:
            Description if found else empty string

        """

        # There may be multiple
        matched_details = self.lookup_column_output(frame, column, outputs)

        for detail in matched_details:

            if not isinstance(detail, dict):
                # Skip due to invalid format.
                continue

            generator = detail.get("generator", None)

            if generator is None:
                # Nothing to parse
                return detail.get("description", "")

            # Now parse and apply to generate
            variables = detail.get("variables", [])
            context = {}
            for v in variables:
                try:
                    context[v] = matcher.group(v)
                except:
                    context[v] = "error"

            try:
                return generator % context
            except:
                logger.error("Error in configuration", extra={"transform": self.name})
                pass

        return ""

    def get_column_security(self, frame, column, outputs=None):
        """
        Look through the output definition to extract the security
        attributes including privacy.

        Args:
            frame (object): name of the dataframe being processed
            column (object): Name of the column whose security attributes are required

        Returns:
            dict with attribute. Could be multiple

        """
        # There may be multiple matches
        matched_details = self.lookup_column_output(frame, column, outputs)
        if len(matched_details) > 0:
            return matched_details[0].get("security", {})

        return {}

    def get_column_version(self, frame, column, outputs=None):
        """
        Look through the output definition to extract version  attributes

        Args:
            frame (object): name of the dataframe being processed
            column (object): Name of the column whose version attributes are required

        Returns:
            string (version)

        """
        # There may be multiple matches
        matched_details = self.lookup_column_output(frame, column, outputs)
        if len(matched_details) > 0:
            return matched_details[0].get("version", "v1")

        return "v1"  # default

    def validate_results(self, what, state):
        """
        Validate outputs generated by the transform
        """
        pass


class Model(Transform):
    """
    Transforms that are models.

    As of now no additional function interfaces exist. But model
    management functions will be added in future.

    """

    def __init__(self, *args, **kwargs):
        super(Model, self).__init__(*args, **kwargs)
        self.roles_supported = ["Compute"]
        self.roles_current = "Compute"


##################################################
# Integration
##################################################
class Integration(Node):
    """
    Integration with thirdparty service
    """

    def __init__(self, *args, **kwargs):
        super(Integration, self).__init__(*args, **kwargs)
        self.node_type = "Integration"



import os
import sys
import json
import re
import copy
from enrichsdk import Sink
import logging

logger = logging.getLogger("app")


class JSONSink(Sink):
    """
    Store a 'dict' frame that is present in the state into a file.

    Params are meant to be passed as parameter to update_frame.

    Example configuration::

         "args": {
             "sink": {
                 'test': {
                     'frametype': 'dict',
                     'filename': '%(output)s/%(runid)s/mytestoutput.json',
                     'params': {}
                 }
             }
         }
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "JSONSink"

        self.testdata = {
            "conf": {
                "args": {
                    "sink": {
                        "frame1": {
                            "frametype": "dict",
                            "filename": "%(output)s/%(runid)s/mytestoutput.json",
                            "params": {},
                        }
                    }
                }
            },
            "data": {
                "frame1": {
                    "filename": "outputjson.json",
                    "frametype": "dict",
                    "transform": "TestJSON",
                    "params": {},
                }
            },
        }

    def preload_clean_args(self, args):
        """
        Clean when the spec is loaded...
        """

        if "sink" not in args:
            args = {"sink": args}

        args = super().preload_clean_args(args)

        assert "sink" in args
        assert isinstance(args["sink"], dict)
        assert len(args["sink"]) > 0

        sink = args["sink"]
        for name, detail in sink.items():

            if ("frametype" not in detail) or (detail["frametype"] != "dict"):
                logger.error(
                    "Invalid configuration. Only JSON/Dictionaries are supported by this sink transform",
                    extra=self.config.get_extra({"transform": self.name}),
                )
                raise Exception("Invalid configuration")

            if (
                ("filename" not in detail)
                or (not isinstance(detail["filename"], str))
                or ("params" not in detail)
                or (not isinstance(detail["params"], dict))
            ):
                logger.error(
                    "Invalid args. Filename (string) and params (dict) are required",
                    extra=self.config.get_extra({"transform": self.name}),
                )
                raise Exception("Invalid configuration")

            filename = detail["filename"]
            if not filename.lower().endswith(".json"):
                logger.error(
                    "Input file must a .json file",
                    extra=self.config.get_extra({"transform": self.name}),
                )
                raise Exception("Invalid configuration")

            detail["root"] = self.config.enrich_data_dir

            tags = detail.get("tags", [])
            if isinstance(tags, str):
                tags = [tags]
            detail["tags"] = tags

            # => Materialize the path...
            detail["filename"] = self.config.get_file(
                detail["filename"], extra={"frame_name": name}
            )

        return args

    def validate_args(self, what, state):
        """
        An extra check on the arguments to make sure
        it is consistent with the specification
        """
        args = self.args
        assert "sink" in args
        assert isinstance(args["sink"], dict)
        assert len(args["sink"]) > 0

        sink = args["sink"]
        for name, detail in sink.items():
            assert ("frametype" in detail) and (detail["frametype"] == "dict")
            assert "filename" in detail
            assert "params" in detail

    def process(self, state):
        """
        Store the dictionary 'frames' in the state in files.
        """
        logger.debug(
            "{} - process".format(self.name),
            extra=self.config.get_extra({"transform": self.name}),
        )

        available_frames = state.get_frame_list()

        # => First construct input for the pandasframe
        extra = {}
        args_input = {}
        write_input = {}
        framecls = self.config.get_dataframe("dict")

        sink = self.args["sink"]
        for pattern in sink:
            # The pattern could be precise dataframe name or could be
            # regular expression.
            regex = re.compile("^{}$".format(pattern))
            frames = [
                m.group(0) for f in available_frames for m in [regex.search(f)] if m
            ]
            if len(frames) == 0:
                logger.warning("Pattern has not matched any frames: {}".format(pattern))
                continue

            for f in frames:

                # Get the details of this frame
                detail = state.get_frame(f)

                # Handle frametype
                frametype = detail["frametype"]
                if frametype != "dict":
                    logger.warning(
                        "Pattern has matched non-dict frame: {}".format(f),
                        extra=self.config.get_extra({"transform": self.name}),
                    )
                    continue

                # Now construct the output file name
                filename = sink[pattern]["filename"]
                filename = self.config.get_file(
                    filename, create_dir=True, extra={"frame_name": f}
                )

                extra[f] = {
                    "notes": self.collapse_notes(detail),
                    "descriptions": self.collapse_descriptions(detail),
                }

                params = sink[pattern].get("params", {})
                write_input[f] = {
                    "frametype": detail["frametype"],
                    "filename": filename,
                    "pattern": pattern,
                    "df": detail["df"],
                    "params": params,
                }

                args_input[f] = copy.copy(sink[pattern])
                args_input[f]["filename"] = filename

        framecls.write(args_input, write_input)

        for name in write_input:

            detail = write_input[name]

            # => Insert columns and tags
            pattern = detail["pattern"]
            detail["params"]["tags"] = sink[pattern]["tags"]

            # Incorporate columns, notes and description
            detail["params"].update(extra[name])

            detail["params"] = [
                detail["params"],
                {
                    "type": "lineage",
                    "transform": self.name,
                    "dependencies": [
                        {"type": "dataframe", "nature": "input", "objects": [name]},
                        {
                            "type": "file",
                            "nature": "output",
                            "objects": [detail["params"]["filename"]],
                        },
                    ],
                },
            ]

            # Insert additional detail
            detail["transform"] = self.name
            detail["history"] = [{"transform": self.name, "log": "Wrote output"}]

            state.update_frame(name, detail)

        logger.debug(
            "Finished writing data",
            extra=self.config.get_extra({"transform": self.name}),
        )

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        """
        Check to make sure that the execution completed correctly
        """
        pass


provider = JSONSink



"""
Expectations
^^^^^^^^^^^^

Base classes for expectation and manager, and
builtin expectations.

"""

from .exceptions import *

################################
# Helper
################################
def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)]
    )


################################
# Main Classes
################################
class ExpectationResultBase(object):
    """
    Class to return one or more results from an
    expectation validation step.
    """

    def __init__(self, *args, **kwargs):
        self.results = []

    def add_result(self, expectation, description, passed, extra={}):
        """
        Add an entry to the validation result object

        Args:
           expectation: Name of the expectation
           description: One line summary of the check
           passed: True if the check has passed
           extra: Dictionary with extra context including reason

        """
        if not isinstance(passed, bool):
            raise IncorrectImplementationExpectation(
                "Expectation requires 'passed' to be boolean"
            )

        if not isinstance(expectation, str):
            raise IncorrectImplementationExpectation(
                "Expectation require name to be string"
            )

        if not isinstance(description, str):
            raise IncorrectImplementationExpectation(
                "Expectation require description to be string"
            )

        if not isinstance(extra, dict):
            raise IncorrectImplementationExpectation(
                "Expectation require 'extra' to be dict"
            )

        entry = {
            "expectation": expectation,
            "description": description,
            "passed": passed,
            "extra": extra,
        }

        self.results.append(entry)

    def get_results(self):
        for r in self.results:
            yield r


class ExpectationBase(object):
    """
    Base class for expectations
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the base class for expectations
        """
        self.name = "unknown_table_check"
        self.description = "Unknown description"

    def match(self, config):
        """
        Check if the expectations matches

        Args:
           config: Expectation specification

        Returns:
           True if this class can handle expectation

        """
        if not isinstance(config, dict):
            return False

        return config.get("expectation", None) == self.name

    def args_validator_attrs_list(self, config, attrs):
        params = config.get("params", {})
        for attr in attrs:
            if attr not in params:
                raise InvalidExpectation("Missing parameter: {}".format(attr))

    def validate_args(self, config):
        """
        Validate arguments passed to this expectation

        Check if the expectation is correctly specified beyond the
        name

        Args:
           config: Expectation specification

        Returns:
           True if this class can handle this specification

        """
        pass

    def generate(self, df):
        return []

    def validate(self, df, config):

        result = ExpectationResultBase()
        result.add_result(
            expectation=self.name, description=self.description, passed=True
        )
        return result


# Output of GreatExpectations
# {
#    "expectation_type": "expect_table_row_count_to_be_between",
#    "kwargs": {
#        "min_value": 0,
#        "max_value": None
#    },
#    "meta": {
#        "BasicDatasetProfiler": {
#            "confidence": "very low"
#        }
#    }
# }


class ExpectationManagerBase(object):
    def __init__(self):
        self.subclasses = None

    def initialize(self):
        subclasses = all_subclasses(ExpectationBase)
        self.subclasses = [cls() for cls in subclasses]

    def validate(self, df, expectations):

        if self.subclasses is None:
            raise NotInitialized("Manager not initialized")

        if (not isinstance(expectations, list)) and (
            not isinstance(expectations, dict)
        ):
            raise InvalidExpectation("Expectations should be a list or dict")

        if isinstance(expectations, dict):
            expectations = [expectations]

        if len(expectations) == 0:
            raise NoExpectations("Empty list")

        for e in expectations:
            if (not isinstance(e, dict)) or ("expectation" not in e):
                raise InvalidExpectation("Invalid expectation format")

        result = []
        for e in expectations:
            nomatches = True
            name = e["expectation"]
            for cls in self.subclasses:
                if cls.match(e):
                    cls.validate_args(e)

                    nomatches = False

                    # Now run the validation
                    expectationresult = cls.validate(df, e)

                    if (expectationresult is None) or (
                        not isinstance(expectationresult, ExpectationResultBase)
                    ):
                        raise IncorrectImplementationExpectation(
                            "Expectation returned None or unexpected object"
                        )

                    # Now all the validation results should be added to the common list.
                    result.extend(list(expectationresult.get_results()))

            if nomatches:
                raise UnsupportedExpectation("No handler for {}".format(name))

        return result



import logging
from logging.config import dictConfig

logconfig = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "basic": {
            "class": "logging.Formatter",
            "datefmt": "%I:%M:%S",
            "format": "%(asctime)s %(levelname)s %(name)s %(message)s",
        },
        "expanded": {
            "class": "logging.Formatter",
            "format": "%(asctime)s - %(name)s - %(transform)s - %(levelname)s - %(message)s%(data)s",
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "basic",
            "level": "INFO",
            "stream": "ext://sys.stdout",
        },
        "app": {
            "class": "logging.StreamHandler",
            "formatter": "expanded",
            "level": "DEBUG",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "": {"level": "INFO", "handlers": ["console"], "propagate": False},
        "default": {"level": "INFO", "handlers": ["console"], "propagate": False},
        "app": {"level": "DEBUG", "handlers": ["app"], "propagate": False},
    },
    "root": {"level": "WARN", "handlers": ["console"]},
}


def setup_logging():
    dictConfig(logconfig)

    for name in [
        "kazoo.client",
        "pykafka.simpleconsumer",
        "pykafka.handlers",
        "pykafka.balancedconsumer",
        "pykafka.connection",
        "pykafka.membershipprotocol",
        "pykafka.cluster",
        "pykafka.topic",
        "matplotlib.pyplot",
        "asyncio",
    ]:
        logging.getLogger(name).disabled = True
